# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from lightworks import State
from lightworks.emulator import Source
from lightworks.emulator.state import AnnotatedState


class TestSource:
    """
    Unit tests to check behaviour and statistics returned while using the
    Source object of the emulator.
    """

    def test_perfect_input(self):
        """Checks only one state is returned when a perfect input is used."""
        source = Source()
        stats = source._build_statistics(State([1, 0, 1, 0]))
        assert stats == {State([1, 0, 1, 0]): 1}

    def test_imperfect_brightness(self):
        """Checks the return values when an imperfect brightness is used."""
        source = Source(brightness=0.915)
        stats = source._build_statistics(State([1, 0, 1, 0]))
        stats = {s: round(p, 6) for s, p in stats.items()}
        expected = {
            State([1, 0, 1, 0]): 0.837225,
            State([1, 0, 0, 0]): 0.077775,
            State([0, 0, 1, 0]): 0.077775,
            State([0, 0, 0, 0]): 0.007225,
        }
        assert stats == expected

    def test_imperfect_source(self):
        """
        Checks return values are correct when all possible imperfections in a
        source are used. Also, confirms all states returned are of the correct
        length.
        """
        source = Source(
            brightness=0.493,
            purity=0.984,
            indistinguishability=0.937,
            probability_threshold=0,
        )
        stats = source._build_statistics(State([0, 0, 2, 1, 0, 1, 0, 0]))
        # Check number of inputs is correct
        assert len(stats) == 204
        # Check some random values
        assert stats[
            AnnotatedState([[], [], [0], [0], [], [0], [], []])
        ] == pytest.approx(0.10844531898317525, 1e-6)
        assert stats[AnnotatedState([[]] * 8)] == pytest.approx(
            0.06502113537379095, 1e-6
        )
        # Ensure all state lengths are correct
        for state in stats:
            assert len(state) == 8

    def test_source_thresholding(self):
        """
        Confirms correct behaviour when probability thresholding is applied to
        the Source object.
        """
        source = Source(
            brightness=0.493,
            purity=0.984,
            indistinguishability=0.937,
            probability_threshold=1e-6,
        )
        stats = source._build_statistics(State([0, 0, 2, 1, 0, 1, 0, 0]))
        # Check length is as expected after thresholding
        assert len(stats) == 89

    @pytest.mark.parametrize(
        ("value", "error"),
        [
            (True, TypeError),
            ("0.7", TypeError),
            (0.5, ValueError),
            (1.1, ValueError),
        ],
    )
    def test_purity_modification(self, value, error):
        """
        Checks that the purity cannot be assigned to an invalid value.
        """
        source = Source()
        with pytest.raises(error):
            source.purity = value

    @pytest.mark.parametrize(
        ("value", "error"),
        [
            (True, TypeError),
            ("0.5", TypeError),
            (-0.1, ValueError),
            (1.1, ValueError),
        ],
    )
    def test_brightness_modification(self, value, error):
        """
        Checks that the brightness cannot be assigned to an invalid value.
        """
        source = Source()
        with pytest.raises(error):
            source.brightness = value

    @pytest.mark.parametrize(
        ("value", "error"),
        [
            (True, TypeError),
            ("0.5", TypeError),
            (-0.1, ValueError),
            (1.1, ValueError),
        ],
    )
    def test_indistinguishability_modification(self, value, error):
        """
        Checks that the indistinguishability cannot be assigned to an invalid
        value.
        """
        source = Source()
        with pytest.raises(error):
            source.indistinguishability = value

    @pytest.mark.parametrize(
        ("value", "error"),
        [
            (True, TypeError),
            ("0.5", TypeError),
            (-0.1, ValueError),
            (1.1, ValueError),
        ],
    )
    def test_probability_threshold_modification(self, value, error):
        """
        Checks that the probability_threshold cannot be assigned to an invalid
        value.
        """
        source = Source()
        with pytest.raises(error):
            source.probability_threshold = value



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import copy
from typing import Any

import numpy as np
from numpy.typing import NDArray

from lightworks.sdk.utils.exceptions import ModeRangeError

from .photonic_components import Component, Group, HeraldData, Loss


class CompiledPhotonicCircuit:
    """
    Builds up the unitary representation of a particular circuit, combining
    provided components from a circuit spec.

    Args:

        n_modes (int) : The number of modes in a circuit. This should not
            include any required loss modes.

    """

    def __init__(self, n_modes: int) -> None:
        self._n_modes = n_modes
        self._loss_modes = 0
        self._unitary = np.identity(n_modes, dtype=complex)
        self._in_heralds: dict[int, int] = {}
        self._out_heralds: dict[int, int] = {}
        self._circuit_spec: list[tuple[str, dict[str, Any]] | None] = []

    @property
    def n_modes(self) -> int:
        """The number of modes in the circuit."""
        return self._n_modes

    @property
    def loss_modes(self) -> int:
        """The current number of loss modes in the circuit."""
        return self._loss_modes

    @property
    def total_modes(self) -> int:
        """The total number of modes, including real and loss modes."""
        return self.n_modes + self.loss_modes

    @property
    def U_full(self) -> NDArray[np.complex128]:  # noqa: N802
        """Full unitary matrix."""
        return self._unitary

    @property
    def heralds(self) -> HeraldData:
        """Returns details of heralds on the input and output."""
        return HeraldData(
            input=copy(self._in_heralds),
            output=copy(self._out_heralds),
        )

    @property
    def input_modes(self) -> int:
        """
        The number of input modes that should be specified, accounting for the
        heralds used in the circuit.
        """
        return self.n_modes - len(self.heralds.input)

    def add(self, spec: Component) -> None:
        """Adds elements to the spec."""
        if isinstance(spec, Loss):
            self._loss_modes += 1
            self._unitary = np.pad(
                self._unitary, (0, 1), "constant", constant_values=0j
            )
            self._unitary[-1, -1] = 1 + 0j
            self._circuit_spec.append(spec.serialize())
        if isinstance(spec, Group):
            for s in spec.circuit_spec:
                self.add(s)
        else:
            self._unitary = spec.get_unitary(self.total_modes) @ self._unitary
            self._circuit_spec.append(spec.serialize())

    def add_herald(
        self, input_mode: int, output_mode: int, in_photon: int, out_photon: int
    ) -> None:
        """
        Add a herald across a selected input/output of the circuit. If only one
        mode is specified then this will be used for both the input and output.

        Args:

            input_mode (int) : The input mode to use for the herald.

            output_mode (int) : The output mode to use for the herald.

            in_photon (int) : The number of photons to use on the heralding
                input.

            out_photon (int) : The number of photons to use on the heralding
                output.

        """
        for n in [in_photon, out_photon]:
            if not isinstance(n, int) or isinstance(n, bool):
                raise TypeError(
                    "Number of photons for herald should be an integer."
                )
        in_photon = int(in_photon)
        out_photon = int(out_photon)
        if output_mode is None:
            output_mode = input_mode
        self._mode_in_range(input_mode)
        self._mode_in_range(output_mode)
        # Check if herald already used on input or output
        if input_mode in self._in_heralds:
            raise ValueError("Heralding already set for chosen input mode.")
        if output_mode in self._out_heralds:
            raise ValueError("Heralding already set for chosen output mode.")
        # If not then update dictionaries
        self._in_heralds[input_mode] = in_photon
        self._out_heralds[output_mode] = out_photon

    def _mode_in_range(self, mode: int) -> None:
        """
        Check a mode exists within the created circuit and also confirm it
        is an integer.
        """
        if not isinstance(mode, int) or isinstance(mode, bool):
            raise TypeError("Mode number should be an integer.")
        if not 0 <= mode < self.n_modes:
            raise ModeRangeError(
                "Selected mode(s) is not within the range of the created "
                "circuit."
            )



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from lightworks.sdk.circuit import Parameter, PhotonicCircuit

from .task import Task


class Batch:
    """
    Allows for batch submission of jobs with varied parameters on each run.
    """

    def __init__(
        self,
        task: type[Task] | None = None,
        task_args: list[list[Any]] | None = None,
        task_kwargs: dict[str, list[Any]] | None = None,
        parameters: dict[Parameter[Any], list[Any]] | None = None,
    ) -> None:
        # Define list to store tasks
        self.__tasks: list[Task] = []
        if task is None:
            return  # Skip all setup if no task object is provided
        # Check task not already initialized.
        if not isinstance(task, type):
            raise TypeError("Provided task should not be initialized.")
        args, kwargs, params, n_vars = process_values(
            task_args, task_kwargs, parameters
        )
        # Loop over number of variations and create a task for each
        for i in range(n_vars):
            if params is not None:
                for p, v in params.items():
                    p.set(v[i])
            sub_args = (
                [copy_circuit(a[i]) for a in args] if args is not None else []
            )
            sub_kwargs = (
                {k: copy_circuit(v[i]) for k, v in kwargs.items()}
                if kwargs is not None
                else {}
            )
            try:
                self.__tasks.append(task(*sub_args, **sub_kwargs))
            except Exception as e:
                msg = (
                    "An exception occurred while trying to create task number "
                    f"{i + 1}. Please see the above exception for more details."
                )
                raise ValueError(msg) from e

    def __iter__(self) -> Any:
        """Iterable to allow to loop over each task."""
        yield from self.tasks

    @property
    def tasks(self) -> list[Task]:
        """Returns all tasks contained in batch."""
        return self.__tasks

    @property
    def num(self) -> int:
        """Returns number of tasks in batch."""
        return len(self.tasks)

    def add(self, task: Task) -> None:
        """Allows for tasks to be manually added to the batch."""
        if not isinstance(task, Task):
            raise TypeError("Provided task must be a Task object.")
        self.__tasks.append(task)


def copy_circuit(arg: Any) -> Any:
    """
    Checks if a provided argument is a circuit and if so makes a copy and freeze
    the circuit parameters.
    """
    if isinstance(arg, PhotonicCircuit):
        return arg.copy(freeze_parameters=True)
    return arg


def process_values(
    task_args: list[list[Any]] | None,
    task_kwargs: dict[str, list[Any]] | None,
    parameters: dict[Parameter[Any], list[Any]] | None,
) -> tuple[
    list[Any] | None,
    dict[str, list[Any]] | None,
    dict[Parameter[Any], list[Any]] | None,
    int,
]:
    """
    Processes args, kwargs and parameter values passed to a BatchTask.
    """
    n_vars = 0
    # Confirm each entry is list or tuple
    if task_args is not None:
        for a in task_args:
            if not isinstance(a, list | tuple):
                raise TypeError(
                    "Each value for task_args should be provided as a list."
                )
        n_vars = max([len(a) for a in task_args] + [n_vars])
    if task_kwargs is not None:
        for k, v in task_kwargs.items():
            if not isinstance(k, str):
                raise TypeError("'task_kwargs' keys should be strings.")
            if not isinstance(v, list | tuple):
                raise TypeError(
                    "Each value for task_kwargs should be provided as a list."
                )
        n_vars = max([len(a) for a in task_kwargs.values()] + [n_vars])
    if parameters is not None:
        for p, v in parameters.items():
            if not isinstance(p, Parameter):
                raise TypeError(
                    "'parameters' keys should be Parameter objects."
                )
            if not isinstance(v, list | tuple):
                raise TypeError(
                    "Each value for parameters should be provided as a list."
                )
        n_vars = max([len(a) for a in parameters.values()] + [n_vars])
    # Then check number of values is of the correct length
    if task_args is not None:
        for i, a in enumerate(task_args):
            if len(a) == 1:
                task_args[i] = a * n_vars
            elif len(a) != n_vars:
                raise ValueError(
                    "Mismatch in number of values provided for one of the "
                    "task_args."
                )
    if task_kwargs is not None:
        for k, v in task_kwargs.items():
            if len(v) == 1:
                task_kwargs[k] = v * n_vars
            elif len(v) != n_vars:
                raise ValueError(
                    "Mismatch in number of values provided for one of the "
                    "task_kwargs."
                )
    if parameters is not None:
        for p, v in parameters.items():
            if len(v) == 1:
                parameters[p] = v * n_vars
            elif len(v) != n_vars:
                raise ValueError(
                    "Mismatch in number of values provided for one of the "
                    "parameters."
                )
    return (task_args, task_kwargs, parameters, n_vars)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from random import random

import pytest

from lightworks import (
    Analyzer,
    Batch,
    Parameter,
    PhotonicCircuit,
    PostSelection,
    Sampler,
    Simulator,
    State,
    Unitary,
    random_unitary,
)
from lightworks.emulator import Backend


class TestBatchTask:
    """
    Unit tests for the Batch object.
    """

    def setup_method(self):
        """
        Creates components required for testing
        """
        self.circuits = [Unitary(random_unitary(8)) for _i in range(4)]
        self.inputs = [
            State([1, 1, 0, 0, 1, 1, 0, 0]),
            State([1, 1, 0, 0, 0, 0, 1, 1]),
            State([0, 0, 1, 1, 1, 1, 0, 0]),
            State([0, 0, 1, 1, 0, 0, 1, 1]),
        ]
        self.samples = [1, 100, 10000, 1000000]
        self.min_detections = [1, 2, 3, 4]
        self.post_selection = [PostSelection()] * 4

    def test_task_setup(self):
        """
        Checks basic Batch setup works correctly.
        """
        Batch(
            Sampler,
            task_args=[self.circuits, self.inputs, self.samples],
            task_kwargs={
                "min_detection": self.min_detections,
                "post_selection": self.post_selection,
            },
        )

    def test_setup_no_task(self):
        """
        Checks Batch can be initialized with no arguments.
        """
        Batch()

    def test_setup_task_initialised(self):
        """
        Checks an error is raised when an already initialized task is provided
        to Batch.
        """
        sim = Simulator(PhotonicCircuit(2), State([1, 0]))
        with pytest.raises(TypeError):
            Batch(sim)

    def test_task_setup_all_kwargs(self):
        """
        Checks basic Batch setup works correctly when all arguments are
        specified as keyword arguments.
        """
        Batch(
            Sampler,
            task_kwargs={
                "circuit": self.circuits,
                "input_state": self.inputs,
                "n_samples": self.samples,
                "min_detection": self.min_detections,
                "post_selection": self.post_selection,
            },
        )

    def test_task_setup_single_value(self):
        """
        Checks basic Batch setup works correctly when only a single value
        is provided for some of the args.
        """
        Batch(
            Sampler,
            task_args=[[self.circuits[1]], [self.inputs[2]], self.samples],
            task_kwargs={
                "min_detection": [self.min_detections[3]],
                "post_selection": [self.post_selection[0]],
            },
        )

    def test_task_number(self):
        """
        Checks that the correct number of tasks is generated by the
        Batch.
        """
        bc = Batch(
            Sampler,
            task_args=[self.circuits, self.inputs, self.samples],
            task_kwargs={"min_detection": self.min_detections},
        )
        assert bc.num == 4

    def test_add(self):
        """
        Checks add function allows a task to be added, and raised an error for
        not-tasks.
        """
        sim = Simulator(PhotonicCircuit(2), State([1, 0]))
        batch = Batch()
        batch.add(sim)
        assert batch.num == 1

    def test_add_not_task(self):
        """
        Checks an error is raised when a non-task is added to Batch.
        """
        batch = Batch()
        with pytest.raises(TypeError):
            batch.add(PhotonicCircuit(4))

    def test_task_number_all_kwargs(self):
        """
        Checks that the correct number of tasks is generated by the
        Batch when all args are specified by keyword arguments.
        """
        bc = Batch(
            Sampler,
            task_kwargs={
                "circuit": self.circuits,
                "input_state": self.inputs,
                "n_samples": self.samples,
                "min_detection": self.min_detections,
                "post_selection": self.post_selection,
            },
        )
        assert bc.num == 4

    def test_task_values(self):
        """
        Checks correct values are set in each of the tasks generated.
        """
        bc = Batch(
            Sampler,
            task_args=[self.circuits, self.inputs, self.samples],
            task_kwargs={
                "min_detection": self.min_detections,
                "post_selection": self.post_selection,
            },
        )
        for i, task in enumerate(bc.tasks):
            assert (task.circuit.U_full == self.circuits[i].U_full).all()
            assert task.input_state == self.inputs[i]
            assert task.n_samples == self.samples[i]
            assert task.min_detection == self.min_detections[i]
            assert task.post_selection == self.post_selection[i]

    def test_task_values_all_kwargs(self):
        """
        Checks correct values are set in each of the tasks generated when all
        args are provided as keyword arguments.
        """
        bc = Batch(
            Sampler,
            task_kwargs={
                "circuit": self.circuits,
                "input_state": self.inputs,
                "n_samples": self.samples,
                "min_detection": self.min_detections,
                "post_selection": self.post_selection,
            },
        )
        for i, task in enumerate(bc.tasks):
            assert (task.circuit.U_full == self.circuits[i].U_full).all()
            assert task.input_state == self.inputs[i]
            assert task.n_samples == self.samples[i]
            assert task.min_detection == self.min_detections[i]
            assert task.post_selection == self.post_selection[i]

    def test_task_with_parameter_setup(self):
        """
        Checks that parameters can be provided as part of the setup of the
        Batch.
        """
        parameters = {Parameter(1): [1, 2, 3, 4], Parameter(1): [5, 6, 7, 8]}
        Batch(
            Sampler,
            task_kwargs={
                "circuit": self.circuits,
                "input_state": self.inputs,
                "n_samples": self.samples,
                "min_detection": self.min_detections,
                "post_selection": self.post_selection,
            },
            parameters=parameters,
        )

    def test_task_creates_different_circuits(self):
        """
        Checks that parameters can be provided as part of the setup of the
        Batch and that this creates unique circuits.
        """
        # Generate parametrised circuit
        parameters = {}
        circuit = PhotonicCircuit(8)
        for i, m in enumerate([0, 2, 4, 6, 1, 3, 5, 0, 2, 4, 6]):
            theta = Parameter(1, label=f"theta_{i}")
            phi = Parameter(1, label=f"phi_{i}")
            circuit.ps(m, phi)
            circuit.bs(m)
            circuit.ps(m + 1, theta)
            circuit.bs(m)
            parameters[theta] = [6.3 * random() for i in range(4)]
            parameters[phi] = [6.3 * random() for i in range(4)]
        # Then setup batch task
        bc = Batch(
            Sampler,
            task_kwargs={
                "circuit": [circuit],
                "input_state": self.inputs,
                "n_samples": self.samples,
                "min_detection": self.min_detections,
                "post_selection": self.post_selection,
            },
            parameters=parameters,
        )
        # deepcopy tasks to ensure no modification by parameters
        all_tasks = deepcopy(bc.tasks)
        # Check all unitaries are equivalent to the expected from the circuit
        for i, task in enumerate(all_tasks):
            for p in parameters:
                p.set(parameters[p][i])
            assert (task.circuit.U.round() == circuit.U.round()).all()

    @pytest.mark.parametrize(
        "outputs",
        [
            None,
            [None],
            [State([0, 1, 0, 1])],
            [State([0, 1, 0, 1])] * 5,
            [[State([0, 1, 0, 1]), State([0, 1, 1, 0])]] * 5,
        ],
    )
    def test_simulator_batch(self, outputs):
        """
        Checks the simulator can be run as a batch.
        """
        circuits = [Unitary(random_unitary(4)) for _ in range(5)]
        inputs = [State([1, 0, 1, 0])]
        kwargs = {"outputs": outputs} if outputs is not None else {}
        batch = Batch(Simulator, [circuits, inputs], kwargs)
        Backend("permanent").run(batch)

    @pytest.mark.parametrize(
        "expected",
        [None, [None], [{State([1, 0, 1, 0]): State([1, 0, 1, 0])}]],
    )
    def test_analyzer_batch(self, expected):
        """
        Checks the simulator can be run as a batch.
        """
        circuits = [Unitary(random_unitary(4)) for _ in range(5)]
        inputs = [State([1, 0, 1, 0])]
        kwargs = {"expected": expected} if expected is not None else {}
        batch = Batch(Analyzer, [circuits, inputs], kwargs)
        Backend("permanent").run(batch)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any, TypeVar

from lightworks.sdk.state import State
from lightworks.sdk.utils.exceptions import ResultError
from lightworks.sdk.utils.post_selection import PostSelectionType

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")

RT = TypeVar("RT")


class Result(dict[_KT, _VT], ABC):
    """
    Base class for all Lightworks result objects.
    """

    @abstractmethod
    def map(
        self: RT,
        mapping: Callable[[State, Any], State],
        *args: Any,
        **kwargs: Any,
    ) -> RT:
        """
        Requires ability to apply a state mapping on a particular result.
        """

    @abstractmethod
    def apply_post_selection(
        self: RT, post_selection: PostSelectionType | Callable[[State], bool]
    ) -> RT:
        """
        Requite the ability to provided additional post-selection criteria to a
        result.
        """

    def __setitem__(self, key: _KT, value: _VT) -> None:
        raise ResultError(
            "Result objects should not be modified directly. To get a copy of "
            "the data which can be modified use dict(Result)."
        )



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Lightworks Tomography
=====================

A set of tools for quantum state & process tomography on a system.
"""

from .gate_fidelity import GateFidelity
from .process_tomography_li import LIProcessTomography
from .process_tomography_mle import MLEProcessTomography
from .projection import project_choi_to_physical, project_density_to_physical
from .state_tomography import StateTomography
from .utils import (
    choi_from_unitary,
    density_from_state,
    process_fidelity,
    state_fidelity,
)

__all__ = [
    "GateFidelity",
    "LIProcessTomography",
    "MLEProcessTomography",
    "StateTomography",
    "choi_from_unitary",
    "density_from_state",
    "process_fidelity",
    "project_choi_to_physical",
    "project_density_to_physical",
    "state_fidelity",
]



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from random import random

import numpy as np
import pytest

from lightworks import PhotonicCircuit, random_unitary
from lightworks.interferometers.decomposition import (
    bs_matrix,
    check_null,
    reck_decomposition,
)


class TestDecomposition:
    """
    Tests for decomposition module.
    """

    @pytest.mark.parametrize("n_modes", [2, 7, 8])
    def test_decomposition(self, n_modes):
        """
        Checks decomposition is able to pass successfully for a valid unitary
        matrix.
        """
        unitary = random_unitary(n_modes)
        reck_decomposition(unitary)

    @pytest.mark.parametrize("n_modes", [2, 7, 8])
    def test_decomposition_identity(self, n_modes):
        """
        Checks decomposition is able to pass successfully for an identity
        matrix.
        """
        unitary = np.identity(n_modes, dtype=complex)
        reck_decomposition(unitary)

    @pytest.mark.parametrize("n_modes", [2, 7, 8])
    def test_decomposition_failed(self, n_modes):
        """
        Checks decomposition fails for a non-unitary matrix.
        """
        unitary = np.zeros((n_modes, n_modes), dtype=complex)
        for i in range(n_modes):
            for j in range(n_modes):
                unitary[i, j] = random() + 1j * random()
        with pytest.raises(ValueError):
            reck_decomposition(unitary)

    def test_bs_matrix(self):
        """
        Check beam splitter matrix is correct for the unit cell used.
        """
        theta, phi = 2 * np.pi * random(), 2 * np.pi * random()
        # Get beam splitter matrix
        bs_u = bs_matrix(0, 1, theta, phi, 2)
        # Create unit cell circuit
        circ = PhotonicCircuit(2)
        circ.ps(0, phi)
        circ.bs(0)
        circ.ps(1, theta)
        circ.bs(0)
        circ_u = circ.U
        # Check equivalence
        assert (bs_u.round(8) == circ_u.round(8)).all()

    @pytest.mark.parametrize("n_modes", [2, 7, 8])
    def test_check_null(self, n_modes):
        """
        Checks null matrix returns True for a diagonal matrix.
        """
        unitary = np.identity(n_modes, dtype=complex)
        for i in range(n_modes):
            unitary[i, i] *= np.exp(1j * random())
        assert check_null(unitary)

    def test_check_null_false(self):
        """
        Checks null matrix returns false for a non-nulled matrix.
        """
        unitary = random_unitary(8)
        assert not check_null(unitary)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from random import shuffle

import numpy as np
import pytest

from lightworks import PostSelection, Sampler, emulator, qubit
from lightworks.tomography import (
    GateFidelity,
    LIProcessTomography,
    MLEProcessTomography,
    choi_from_unitary,
)
from lightworks.tomography.experiments import (
    ProcessTomographyExperiment,
    ProcessTomographyList,
)
from lightworks.tomography.process_tomography import _ProcessTomography


def run_experiments(experiments, n_qubits):
    """
    Experiment function for testing process tomography. The number of qubits
    should be specified in experiment_args.
    """
    post_select = PostSelection()
    for i in range(n_qubits):
        post_select.add((2 * i, 2 * i + 1), 1)
    results = []
    backend = emulator.Backend("slos")
    for exp in experiments:
        sampler = Sampler(
            exp.circuit,
            exp.input_state,
            20000,
            post_selection=post_select,
            random_seed=99,
        )
        results.append(backend.run(sampler))
    return results


cnot_mat = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])

h_exp = choi_from_unitary([[2**-0.5, 2**-0.5], [2**-0.5, -(2**-0.5)]])

cnot_exp = choi_from_unitary(cnot_mat)


class TestLIProcessTomography:
    """
    Unit tests for LIProcessTomography routine. This is also used for more
    general testing in instances where datasets are required.
    """

    def setup_class(self):
        """
        Runs process tomography experiments so results can be reused.
        """
        # Hadamard tomography
        n_qubits = 1
        circ = qubit.H()
        experiments = LIProcessTomography(n_qubits, circ).get_experiments()
        self.h_data = run_experiments(experiments, n_qubits)
        # CNOT tomography
        n_qubits = 2
        circ = qubit.CNOT()
        experiments = LIProcessTomography(n_qubits, circ).get_experiments()
        self.cnot_data = run_experiments(experiments, n_qubits)
        self.cnot_data_dict = {
            (exp.input_basis, exp.measurement_basis): d
            for exp, d in zip(experiments, self.cnot_data, strict=True)
        }

    def test_hadamard_choi(self):
        """
        Checks process tomography of the Hadamard gate produces the expected
        choi matrix.
        """
        tomo = LIProcessTomography(1, qubit.H())
        tomo.process(self.h_data)
        assert tomo.choi == pytest.approx(h_exp, abs=5e-2)

    def test_hadamard_fidelity(self):
        """
        Checks fidelity of hadamard gate process matrix is close to 1.
        """
        tomo = LIProcessTomography(1, qubit.H())
        tomo.process(self.h_data)
        assert tomo.fidelity(h_exp) == pytest.approx(1, 1e-2)

    def test_cnot_choi(self):
        """
        Checks process tomography of the CNOT gate produces the expected choi
        matrix and the fidelity is calculated to be 1.
        """
        tomo = LIProcessTomography(2, qubit.CNOT())
        tomo.process(self.cnot_data)
        assert tomo.choi == pytest.approx(cnot_exp, abs=5e-2)

    def test_cnot_fidelity(self):
        """
        Checks fidelity of CNOT gate process matrix is close to 1.
        """
        tomo = LIProcessTomography(2, qubit.CNOT())
        tomo.process(self.cnot_data)
        assert tomo.fidelity(cnot_exp) == pytest.approx(1, 1e-2)

    def test_cnot_state_fidelity(self):
        """
        Checks that the fidelity of the density matrices for each input is close
        to 1.
        """
        tomo = LIProcessTomography(2, qubit.CNOT())
        fidelity = tomo.process_state_fidelities(self.cnot_data, cnot_mat)
        for f in fidelity.values():
            assert f == pytest.approx(1, 1e-3)

    def test_cnot_density_matrices(self):
        """
        Checks that the calculated density matrices for each input is close to
        the expected input.
        """
        tomo = LIProcessTomography(2, qubit.CNOT())
        rhos = tomo.process_density_matrices(self.cnot_data)
        rhos_exp = tomo.get_expected_density_matrices(cnot_mat)
        for i in rhos:
            assert rhos[i] == pytest.approx(rhos_exp[i], abs=5e-2)

    def test_dict_processing(self):
        """
        Checks process tomography is successful when the data is provided as a
        dictionary.
        """
        tomo = LIProcessTomography(2, qubit.CNOT())
        tomo.process(self.cnot_data_dict)
        assert tomo.fidelity(cnot_exp) == pytest.approx(1, 1e-2)
        assert tomo.choi == pytest.approx(cnot_exp, abs=5e-2)

    def test_dict_processing_shuffled(self):
        """
        Checks process tomography is successful when the data is provided as a
        dictionary and has been shuffled.
        """
        tomo = LIProcessTomography(2, qubit.CNOT())
        shuffled_data = {}
        keys = list(self.cnot_data_dict.keys())
        shuffle(keys)
        for k in keys:
            shuffled_data[k] = self.cnot_data_dict[k]
        tomo.process(shuffled_data)
        assert tomo.fidelity(cnot_exp) == pytest.approx(1, 1e-2)
        assert tomo.choi == pytest.approx(cnot_exp, abs=5e-2)


class TestMLEProcessTomography:
    """
    Unit tests for MLEProcessTomography routine.
    """

    def setup_class(self):
        """
        Runs process tomography experiments so results can be reused.
        """
        # Hadamard tomography
        n_qubits = 1
        circ = qubit.H()
        experiments = MLEProcessTomography(n_qubits, circ).get_experiments()
        self.h_data = run_experiments(experiments, n_qubits)
        # CNOT tomography
        n_qubits = 2
        circ = qubit.CNOT()
        experiments = MLEProcessTomography(n_qubits, circ).get_experiments()
        self.cnot_data = run_experiments(experiments, n_qubits)
        self.cnot_data_dict = {
            (exp.input_basis, exp.measurement_basis): d
            for exp, d in zip(experiments, self.cnot_data, strict=True)
        }

    def test_hadamard_choi(self):
        """
        Checks process tomography of the Hadamard gate produces the expected
        choi matrix.
        """
        tomo = MLEProcessTomography(1, qubit.H())
        tomo.process(self.h_data)
        assert tomo.choi == pytest.approx(h_exp, abs=5e-2)

    def test_hadamard_fidelity(self):
        """
        Checks fidelity of hadamard gate process matrix is close to 1.
        """
        tomo = MLEProcessTomography(1, qubit.H())
        tomo.process(self.h_data)
        assert tomo.fidelity(h_exp) == pytest.approx(1, 1e-2)

    def test_cnot_choi(self):
        """
        Checks process tomography of the CNOT gate produces the expected choi
        matrix and the fidelity is calculated to be 1.
        """
        tomo = MLEProcessTomography(2, qubit.CNOT())
        tomo.process(self.cnot_data)
        assert tomo.choi == pytest.approx(cnot_exp, abs=5e-2)

    def test_cnot_fidelity(self):
        """
        Checks fidelity of CNOT gate process matrix is close to 1.
        """
        tomo = MLEProcessTomography(2, qubit.CNOT())
        tomo.process(self.cnot_data)
        assert tomo.fidelity(cnot_exp) == pytest.approx(1, 1e-2)

    def test_dict_processing(self):
        """
        Checks process tomography is successful when the data is provided as a
        dictionary.
        """
        tomo = MLEProcessTomography(2, qubit.CNOT())
        tomo.process(self.cnot_data_dict)
        assert tomo.fidelity(cnot_exp) == pytest.approx(1, 1e-2)
        assert tomo.choi == pytest.approx(cnot_exp, abs=5e-2)

    def test_cnot_state_fidelity(self):
        """
        Checks that the fidelity of the density matrices for each input is close
        to 1. This is checked for MLE tomography as there is a larger number of
        input states.
        """
        tomo = MLEProcessTomography(2, qubit.CNOT())
        fidelity = tomo.process_state_fidelities(self.cnot_data, cnot_mat)
        for f in fidelity.values():
            assert f == pytest.approx(1, 1e-3)

    def test_cnot_density_matrices(self):
        """
        Checks that the calculated density matrices for each input is close to
        the expected input. This is checked for MLE tomography as there is a
        larger number of input states.
        """
        tomo = MLEProcessTomography(2, qubit.CNOT())
        rhos = tomo.process_density_matrices(self.cnot_data)
        rhos_exp = tomo.get_expected_density_matrices(cnot_mat)
        for i in rhos:
            assert rhos[i] == pytest.approx(rhos_exp[i], abs=5e-2)


class TestGateFidelity:
    """
    Unit tests for checking GateFidelity routine.
    """

    def setup_class(self):
        """
        Runs experiments so results can be reused.
        """
        # Hadamard fidelity
        n_qubits = 1
        circ = qubit.H()
        experiments = GateFidelity(n_qubits, circ).get_experiments()
        self.h_data = run_experiments(experiments, n_qubits)
        # CNOT fidelity
        n_qubits = 2
        circ = qubit.CNOT()
        experiments = GateFidelity(n_qubits, circ).get_experiments()
        self.cnot_data = run_experiments(experiments, n_qubits)

    def test_hadamard_fidelity(self):
        """
        Checks fidelity of hadamard gate process is close to 1.
        """
        tomo = GateFidelity(1, qubit.H())
        f = tomo.process(
            self.h_data, [[2**-0.5, 2**-0.5], [2**-0.5, -(2**-0.5)]]
        )
        assert f == pytest.approx(1, 1e-2)

    def test_cnot_fidelity(self):
        """
        Checks fidelity of CNOT gate process is close to 1.
        """
        tomo = GateFidelity(2, qubit.CNOT())
        f = tomo.process(
            self.cnot_data,
            cnot_mat,
        )
        assert f == pytest.approx(1, 1e-2)


class TestGeneralProcessTomography:
    """
    Contains more general units tests for the process tomography algorithms.
    """

    @pytest.mark.parametrize(
        ("tomography", "exp_len"),
        [
            (LIProcessTomography, 144),
            (MLEProcessTomography, 324),
            (GateFidelity, 144),
        ],
    )
    def test_experiments_length(self, tomography, exp_len):
        """
        Checks number of elements returned by ProcessTomography is the correct
        length.
        """
        tomo = tomography(2, qubit.CNOT())
        experiments = tomo.get_experiments()
        assert len(experiments) == exp_len

    @pytest.mark.parametrize(
        "tomography", [LIProcessTomography, MLEProcessTomography, GateFidelity]
    )
    def test_experiments_list(self, tomography):
        """
        Checks that data returned by ProcessTomography get_experiments is a
        ProcessTomographyList.
        """
        tomo = tomography(2, qubit.CNOT())
        experiments = tomo.get_experiments()
        assert isinstance(experiments, ProcessTomographyList)

    @pytest.mark.parametrize(
        "tomography", [LIProcessTomography, MLEProcessTomography, GateFidelity]
    )
    def test_experiments_in_list_are_experiments(self, tomography):
        """
        Checks that data returned by ProcessTomography get_experiments in a list
        of experiments.
        """
        tomo = tomography(2, qubit.CNOT())
        experiments = tomo.get_experiments()
        for exp in experiments:
            assert isinstance(exp, ProcessTomographyExperiment)

    @pytest.mark.parametrize(
        ("n1", "n2"),
        [
            ("all_circuits", "circuit"),
            ("all_inputs", "input_state"),
            ("all_input_basis", "input_basis"),
            ("all_measurement_basis", "measurement_basis"),
        ],
    )
    def test_all_methods(self, n1, n2):
        """
        Checks that the all methods from ProcessTomographyList correctly returns
        the expected list of values in the correct order.
        """
        tomo = _ProcessTomography(2, qubit.CNOT())
        experiments = tomo.get_experiments()
        for exp, quantity in zip(
            experiments, getattr(experiments, n1), strict=True
        ):
            assert getattr(exp, n2) == quantity



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.typing import NDArray

from lightworks import qubit
from lightworks.sdk.circuit import PhotonicCircuit
from lightworks.sdk.state import State

PAULI_MAPPING: dict[str, NDArray[np.complex128]] = {
    "I": np.array([[1, 0], [0, 1]], dtype=np.complex128),
    "X": np.array([[0, 1], [1, 0]], dtype=np.complex128),
    "Y": np.array([[0, -1j], [1j, 0]], dtype=np.complex128),
    "Z": np.array([[1, 0], [0, -1]], dtype=np.complex128),
}

# Pre-calculated density matrices for different quantum states
RHO_MAPPING: dict[str, NDArray[np.complex128]] = {
    "X+": np.array([[1, 1], [1, 1]], dtype=np.complex128) / 2,
    "X-": np.array([[1, -1], [-1, 1]], dtype=np.complex128) / 2,
    "Y+": np.array([[1, -1j], [1j, 1]], dtype=np.complex128) / 2,
    "Y-": np.array([[1, 1j], [-1j, 1]], dtype=np.complex128) / 2,
    "Z+": np.array([[1, 0], [0, 0]], dtype=np.complex128),
    "Z-": np.array([[0, 0], [0, 1]], dtype=np.complex128),
}

# Details the actual input state and transformation required to achieve a target
# input
r_transform = qubit.H()
r_transform.add(qubit.S())
INPUT_MAPPING: dict[str, tuple[State, PhotonicCircuit]] = {
    "X+": (State([1, 0]), qubit.H()),
    "X-": (State([0, 1]), qubit.H()),
    "Y+": (State([1, 0]), r_transform),
    "Y-": (State([0, 1]), r_transform),
    "Z+": (State([1, 0]), qubit.I()),
    "Z-": (State([0, 1]), qubit.I()),
}

# Details transformations required for different measurement types
_y_measure = PhotonicCircuit(2)
_y_measure.add(qubit.S())
_y_measure.add(qubit.Z())
_y_measure.add(qubit.H())
MEASUREMENT_MAPPING: dict[str, PhotonicCircuit] = {
    "X": qubit.H(),
    "Y": _y_measure,
    "Z": qubit.I(),
    "I": qubit.I(),
}



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from random import shuffle

import numpy as np
import pytest

from lightworks import (
    PhotonicCircuit,
    PostSelection,
    Sampler,
    State,
    Unitary,
    qubit,
    random_unitary,
)
from lightworks.emulator import Backend
from lightworks.tomography import StateTomography
from lightworks.tomography.experiments import (
    StateTomographyExperiment,
    StateTomographyList,
)
from lightworks.tomography.state_tomography import MEASUREMENT_MAPPING


def run_experiments(experiments):
    """
    Experiment function with ability to specify the input state used.
    """
    # Find number of qubits using available input modes.
    n_qubits = int(experiments[0].circuit.input_modes / 2)
    n_samples = 25000
    post_selection = PostSelection()
    for i in range(n_qubits):
        post_selection.add((2 * i, 2 * i + 1), 1)
    results = []
    backend = Backend("slos")
    for exp in experiments:
        sampler = Sampler(
            exp.circuit,
            State([1, 0] * n_qubits),
            n_samples,
            post_selection=post_selection,
            random_seed=29,
        )
        results.append(backend.run(sampler))
    return results


class TestStateTomography:
    """
    Unit tests for state tomography class.
    """

    @pytest.mark.parametrize("n_qubits", [1, 2, 3])
    @pytest.mark.filterwarnings("ignore:.*Matrix is ill-conditioned.")
    def test_basic_state(self, n_qubits):
        """
        Checks correct density matrix is produced when performing tomography on
        the |0> X n_qubits state.
        """
        base_circ = PhotonicCircuit(n_qubits * 2)
        tomo = StateTomography(n_qubits, base_circ)
        experiments = tomo.get_experiments()
        data = run_experiments(experiments)
        rho = tomo.process(data)
        rho_exp = np.zeros((2**n_qubits, 2**n_qubits), dtype=complex)
        rho_exp[0, 0] = 1
        assert rho == pytest.approx(rho_exp, abs=1e-2)
        assert tomo.fidelity(rho_exp) == pytest.approx(1, 1e-3)

    @pytest.mark.parametrize("n_qubits", [1, 2, 3])
    def test_ghz_state(self, n_qubits):
        """
        Checks correct density matrix is produced when performing tomography on
        the n_qubit GHZ state.
        """
        base_circ = PhotonicCircuit(n_qubits * 2)
        base_circ.add(qubit.H())
        for i in range(n_qubits - 1):
            base_circ.add(qubit.CNOT(), 2 * i)
        tomo = StateTomography(n_qubits, base_circ)
        experiments = tomo.get_experiments()
        data = run_experiments(experiments)
        rho = tomo.process(data)
        rho_exp = np.zeros((2**n_qubits, 2**n_qubits), dtype=complex)
        rho_exp[0, 0] = 0.5
        rho_exp[0, -1] = 0.5
        rho_exp[-1, 0] = 0.5
        rho_exp[-1, -1] = 0.5
        assert rho == pytest.approx(rho_exp, abs=1e-2)
        assert tomo.fidelity(rho_exp) == pytest.approx(1, 1e-3)

    @pytest.mark.parametrize("to_shuffle", [True, False])
    def test_ghz_state_dict(self, to_shuffle):
        """
        Checks correct density matrix is produced when performing tomography on
        the n_qubit GHZ state and returning results as a dictionary.
        """
        n_qubits = 2
        base_circ = PhotonicCircuit(n_qubits * 2)
        base_circ.add(qubit.H())
        for i in range(n_qubits - 1):
            base_circ.add(qubit.CNOT(), 2 * i)
        tomo = StateTomography(n_qubits, base_circ)
        experiments = tomo.get_experiments()
        data = run_experiments(experiments)
        # Convert data to dict
        data_dict = dict(
            zip(experiments.all_measurement_basis, data, strict=True)
        )
        if to_shuffle:
            keys = list(data_dict.keys())
            shuffle(keys)
            data_dict = {k: data_dict[k] for k in keys}
        rho = tomo.process(data_dict)
        rho_exp = np.zeros((2**n_qubits, 2**n_qubits), dtype=complex)
        rho_exp[0, 0] = 0.5
        rho_exp[0, -1] = 0.5
        rho_exp[-1, 0] = 0.5
        rho_exp[-1, -1] = 0.5
        assert rho == pytest.approx(rho_exp, abs=1e-2)
        assert tomo.fidelity(rho_exp) == pytest.approx(1, 1e-3)

    @pytest.mark.parametrize("n_modes", [2, 3, 5])
    def test_number_of_input_modes_twice_number_of_qubits(self, n_modes):
        """
        Checks that number of input modes must be twice number of qubits,
        corresponding to dual rail encoding.
        """
        with pytest.raises(ValueError):
            StateTomography(2, PhotonicCircuit(n_modes))

    @pytest.mark.parametrize("value", [1.5, "2", None, True])
    def test_n_qubits_must_be_integer(self, value):
        """
        Checks value of n_qubits must be an integer.
        """
        with pytest.raises(TypeError, match="qubits"):
            StateTomography(value, PhotonicCircuit(4))

    @pytest.mark.parametrize(
        "value", [PhotonicCircuit(4).U, [1, 2, 3], None, True]
    )
    def test_base_circuit_must_be_circuit(self, value):
        """
        Checks value of base_circuit must be a PhotonicCircuit object.
        """
        with pytest.raises(TypeError, match="circuit"):
            StateTomography(2, value)

    def test_density_mat_before_calc(self):
        """
        Checks an error is raised if the rho attribute is called before
        tomography is performed.
        """
        tomo = StateTomography(2, PhotonicCircuit(4))
        with pytest.raises(AttributeError):
            tomo.rho  # noqa: B018

    def test_fidleity_before_calc(self):
        """
        Checks an error is raised if a user attempts to calculate fidelity
        before performing tomography.
        """
        tomo = StateTomography(2, PhotonicCircuit(4))
        with pytest.raises(AttributeError):
            tomo.fidelity(np.identity(2))

    def test_base_circuit_unmodified(self):
        """
        Confirms base circuit is unmodified when performing single qubit
        tomography.
        """
        base_circ = PhotonicCircuit(2)
        original_unitary = base_circ.U_full
        tomo = StateTomography(1, base_circ)
        experiments = tomo.get_experiments()
        data = run_experiments(experiments)
        tomo.process(data)
        assert pytest.approx(original_unitary) == base_circ.U

    def test_density_matrix_matches(self):
        """
        Confirms density matrix property returns correct value.
        """
        base_circ = PhotonicCircuit(2)
        tomo = StateTomography(1, base_circ)
        experiments = tomo.get_experiments()
        data = run_experiments(experiments)
        rho1 = tomo.process(data)
        rho2 = tomo.rho
        assert (rho1 == rho2).all()

    @pytest.mark.parametrize("operator", ["I", "X", "Y", "Z"])
    def test_circuit_produces_correct_circuit(self, operator):
        """
        Confirms that create circuit function correctly modifies a base circuit.
        """
        base_circ = Unitary(random_unitary(4))
        op = MEASUREMENT_MAPPING[operator]
        # Get tomography circuit
        tomo = StateTomography(2, base_circ)
        tomo_circ = tomo._create_circuit([MEASUREMENT_MAPPING["I"], op])
        # Modify base circuit and compare
        base_circ.add(MEASUREMENT_MAPPING[operator], 2)
        # Compare
        assert tomo_circ.U_full == pytest.approx(base_circ.U_full)

    def test_circuit_enforces_measurement_length(self):
        """
        Checks that create circuit function will raise an error if the
        measurement string is the wrong length.
        """
        tomo = StateTomography(2, PhotonicCircuit(4))
        with pytest.raises(ValueError):
            tomo._create_circuit("XYZ")

    def test_experiments_length(self):
        """
        Checks number of elements returned by StateTomography is the correct
        length.
        """
        tomo = StateTomography(2, qubit.CNOT())
        experiments = tomo.get_experiments()
        assert len(experiments) == 9

    def test_experiments_list(self):
        """
        Checks that data returned by StateTomography get_experiments is a
        StateTomographyList.
        """
        tomo = StateTomography(2, qubit.CNOT())
        experiments = tomo.get_experiments()
        assert isinstance(experiments, StateTomographyList)

    def test_experiments_in_list_are_experiments(self):
        """
        Checks that data returned by StateTomography get_experiments in a list
        of experiments.
        """
        tomo = StateTomography(2, qubit.CNOT())
        experiments = tomo.get_experiments()
        for exp in experiments:
            assert isinstance(exp, StateTomographyExperiment)

    @pytest.mark.parametrize(
        ("n1", "n2"),
        [
            ("all_circuits", "circuit"),
            ("all_measurement_basis", "measurement_basis"),
        ],
    )
    def test_all_methods(self, n1, n2):
        """
        Checks that the all methods from StateTomographyList correctly returns
        the expected list of values in the correct order.
        """
        tomo = StateTomography(2, qubit.CNOT())
        experiments = tomo.get_experiments()
        for exp, quantity in zip(
            experiments, getattr(experiments, n1), strict=True
        ):
            assert getattr(exp, n2) == quantity



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Counter
from collections.abc import Callable

import numpy as np

from lightworks.emulator.components import Detector, Source
from lightworks.sdk.circuit.photonic_compiler import CompiledPhotonicCircuit
from lightworks.sdk.results import SamplingResult
from lightworks.sdk.state import State
from lightworks.sdk.tasks import SamplerTask
from lightworks.sdk.utils.exceptions import SamplerError
from lightworks.sdk.utils.heralding import (
    add_heralds_to_state,
    remove_heralds_from_state,
)
from lightworks.sdk.utils.post_selection import (
    DefaultPostSelection,
    PostSelectionType,
)
from lightworks.sdk.utils.random import process_random_seed

from .probability_distribution import pdist_calc
from .runner import RunnerABC


class SamplerRunner(RunnerABC):
    """
    Calculates the output probability distribution for a configuration and
    finds produces a set of N samples from this.

    Args:

        data (SamplerTask) : The task which is to be executed.

        pdist_function (Callable) : Function for calculating probability
            distributions for a provided unitary & input.

    Attributes:

        source (Source) : The in-use Source object. If the source in the data
            was originally set to None then this a new default Source object is
            created.

        detector (Detector) : The in-use Detector object. If the detector in the
            data was originally set to None then this a new default Detector
            object is created.

    """

    def __init__(
        self,
        data: SamplerTask,
        pdist_function: Callable[
            [CompiledPhotonicCircuit, State], dict[State, float]
        ],
    ) -> None:
        self.data = data
        self.source = Source() if self.data.source is None else self.data.source
        self.detector = (
            Detector() if self.data.detector is None else self.data.detector
        )
        self.pdist_function = pdist_function

    def distribution_calculator(self) -> dict[State, float]:
        """
        Calculates the output probability distribution for the provided
        configuration. This needs to be done before sampling.
        """
        # Check circuit and input modes match
        if self.data.circuit.input_modes != len(self.data.input_state):
            raise ValueError(
                "Mismatch in number of modes between input and circuit."
            )
        # Add heralds to the included input
        modified_state = add_heralds_to_state(
            self.data.input_state, self.data.circuit.heralds.input
        )
        input_state = State(modified_state)
        # Then build with source
        all_inputs = self.source._build_statistics(input_state)
        # And find probability distribution
        pdist = pdist_calc(self.data.circuit, all_inputs, self.pdist_function)
        # Special case to catch an empty distribution
        if not pdist:
            pdist = {State([0] * self.data.circuit.n_modes): 1}
        # Assign calculated distribution to attribute
        self.probability_distribution = pdist
        herald_modes = list(self.data.circuit.heralds.output.keys())
        self.herald_cache = _HeraldCache(herald_modes)
        return pdist

    def run(self) -> SamplingResult:
        """
        Performs sampling using the calculated probability distribution.

        Returns:

            SamplingResult : A dictionary containing the different output
                states and the number of counts for each one.

        """
        if not hasattr(self, "probability_distribution"):
            raise RuntimeError(
                "Probability distribution has not been calculated. This likely "
                "results from an error in Lightworks."
            )

        post_selection = (
            DefaultPostSelection()
            if self.data.post_selection is None
            else self.data.post_selection
        )
        min_detection = (
            0 if self.data.min_detection is None else self.data.min_detection
        )

        if self.data.sampling_mode == "input":
            return self._sample_N_inputs(
                self.data.n_samples,
                post_selection,
                min_detection,
                self.data.random_seed,
            )
        return self._sample_N_outputs(
            self.data.n_samples,
            post_selection,
            min_detection,
            self.data.random_seed,
        )

    def _sample_N_inputs(  # noqa: N802
        self,
        N: int,  # noqa: N803
        post_select: PostSelectionType,
        min_detection: int = 0,
        seed: int | None = None,
    ) -> SamplingResult:
        """
        Function to sample from the configured system by running N clock cycles
        of the system. In each of these clock cycles the input may differ from
        the target input, dependent on the source properties, and there may be
        a number of imperfections in place which means that photons are not
        measured or false detections occur. This means it is possible to for
        less than N measured states to be returned.

        Args:

            N (int) : The number of samples to take from the circuit.

            post_select (PostSelection) : A PostSelection object or function
                which applies a provided set of post-selection criteria to a
                state.

            min_detection (int, optional) : Post-select on a given minimum
                total number of photons, this should not include any heralded
                photons.

            seed (int|None, optional) : Option to provide a random seed to
                reproducibly generate samples from the function. This is
                optional and can remain as None if this is not required.

        Returns:

            SamplingResult : A dictionary containing the different output
                states and the number of counts for each one.

        """
        try:
            samples = self._gen_samples_from_dist(
                self.probability_distribution, N, seed, normalize=False
            )
        # Sometimes the probability distribution will not quite be normalized,
        # in this case try to re-normalize it.
        except ValueError as e:
            total_p = sum(self.probability_distribution.values())
            if abs(total_p - 1) > 0.01:
                msg = (
                    "Probability distribution significantly deviated from "
                    f"required normalisation ({total_p})."
                )
                raise ValueError(msg) from e
            samples = self._gen_samples_from_dist(
                self.probability_distribution, N, seed, normalize=True
            )
            self.probability_distribution = {
                k: v / total_p for k, v in self.probability_distribution.items()
            }
        filtered_samples = []
        # Get heralds and pre-calculate items
        heralds = self.data.circuit.heralds.output
        if (
            heralds
            and max(heralds.values()) > 1
            and not self.detector.photon_counting
        ):
            raise SamplerError(
                "Non photon number resolving detectors cannot be used when"
                "a heralded mode has more than 1 photon."
            )
        herald_items = list(heralds.items())
        # Set detector seed before sampling
        self.detector._set_random_seed(seed)
        # Process output states
        for state, count in samples.items():
            for _ in range(count):
                output_state = self.detector._get_output(state)
                # Checks herald requirements are met
                for m, n in herald_items:
                    if output_state[m] != n:
                        break
                # If met then remove heralded modes and store
                else:
                    herald_state = (
                        self.herald_cache[output_state]
                        if heralds
                        else output_state
                    )
                    if (
                        post_select.validate(herald_state)
                        and herald_state.n_photons >= min_detection
                    ):
                        filtered_samples.append(herald_state)
        counted = dict(Counter(filtered_samples))
        return SamplingResult(counted, self.data.input_state)

    def _sample_N_outputs(  # noqa: N802
        self,
        N: int,  # noqa: N803
        post_select: PostSelectionType,
        min_detection: int = 0,
        seed: int | None = None,
    ) -> SamplingResult:
        """
        Function to generate N output samples from a system, according to a set
        of selection criteria. The function will raise an error if the
        selection criteria is too strict and removes all outputs. Also note
        this cannot be used to simulate detector dark counts.

        Args:

            N (int) : The number of samples that are to be returned.

            post_select (PostSelection) : A PostSelection object or function
                which applies a provided set of post-selection criteria to a
                state.

            min_detection (int, optional) : Post-select on a given minimum
                total number of photons, this should not include any heralded
                photons.

            seed (int|None, optional) : Option to provide a random seed to
                reproducibly generate samples from the function. This is
                optional and can remain as None if this is not required.

        Returns:

            SamplingResult : A dictionary containing the different output
                states and the number of counts for each one.

        """
        pdist = self.probability_distribution
        if self.detector.p_dark > 0 or self.detector.efficiency < 1:
            raise SamplerError(
                "To use detector dark counts or sub-unity detector efficiency "
                "the sampling mode must be set to 'input'."
            )
        # Get heralds and pre-calculate items
        heralds = self.data.circuit.heralds.output
        if (
            heralds
            and max(heralds.values()) > 1
            and not self.detector.photon_counting
        ):
            raise SamplerError(
                "Non photon number resolving detectors cannot be used when"
                "a heralded mode has more than 1 photon."
            )
        herald_items = list(heralds.items())
        # Convert distribution using provided data
        new_dist: dict[State, float] = {}
        for s, p in pdist.items():
            # Apply threshold detection
            if not self.detector.photon_counting:
                s = State([min(i, 1) for i in s])  # noqa: PLW2901
            # Check heralds
            for m, n in herald_items:
                if s[m] != n:
                    break
            else:
                # Then remove herald modes
                herald_state = self.herald_cache[s] if heralds else s
                # Check state meets min detection and post-selection criteria
                # across remaining modes
                if (
                    herald_state.n_photons >= min_detection
                    and post_select.validate(herald_state)
                ):
                    if herald_state in new_dist:
                        new_dist[herald_state] += p
                    else:
                        new_dist[herald_state] = p
        pdist = new_dist
        # Check some states are found
        if not pdist:
            raise SamplerError(
                "No output states compatible with provided post-selection/"
                "min-detection criteria."
            )
        # Then generate samples and return
        return SamplingResult(
            self._gen_samples_from_dist(pdist, N, seed, normalize=True),
            self.data.input_state,
        )

    def _gen_samples_from_dist(
        self,
        distribution: dict[State, float],
        n_samples: int,
        seed: int | None,
        normalize: bool = False,
    ) -> dict[State, int]:
        """
        Takes a computed probability distribution and generates the request
        number of samples, returning this as a dictionary of states and counts.
        """
        # Re-normalise distribution probabilities
        mapping = dict(enumerate(distribution))
        probs = np.array(list(distribution.values()), dtype=float)
        if normalize:
            probs /= sum(probs)
        # Generate N random samples and then process and count output states
        rng = np.random.default_rng(process_random_seed(seed))
        samples = rng.choice(range(len(mapping)), p=probs, size=n_samples)
        # Count states and convert to results object
        return {mapping[n]: c for n, c in Counter(samples).items()}


class _HeraldCache:
    """
    Used for reducing the number of repeated computations of states with
    heralded modes removed.
    """

    def __init__(self, herald_modes: list[int]) -> None:
        self.cache: dict[State, State] = {}
        self.herald_modes = list(herald_modes)

    def __getitem__(self, state: State) -> State:
        """
        Checks if a state is in the cache, otherwise removes the heralded modes.
        """
        if state not in self.cache:
            self.cache[state] = State(
                remove_heralds_from_state(state, self.herald_modes)
            )
        return self.cache[state]



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .analyzer import Analyzer
from .batch import Batch
from .data import AnalyzerTask, SamplerTask, SimulatorTask, TaskData
from .sampler import Sampler
from .simulator import Simulator
from .task import Task

__all__ = [
    "Analyzer",
    "AnalyzerTask",
    "Batch",
    "Sampler",
    "SamplerTask",
    "Simulator",
    "SimulatorTask",
    "Task",
    "TaskData",
]



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Literal

import drawsvg as draw
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import patches

from lightworks.sdk.utils.exceptions import DisplayError

from .display_utils import MPLSettings, SVGSettings

# ruff: noqa: D102


@dataclass(slots=True, kw_only=True)
class DrawSpec(ABC):
    """Base class for all draw specs."""

    @abstractmethod
    def draw_svg(self) -> draw.DrawingBasicElement | draw.Group: ...

    @abstractmethod
    def draw_mpl(self, axes: plt.Axes) -> None: ...


@dataclass(slots=True, kw_only=True)
class WaveguideDrawing(DrawSpec):
    """
    Contains components for drawing a waveguide using the different
    visualization methods.
    """

    x: float
    y: float
    length: float
    width: float

    def draw_svg(self) -> draw.DrawingBasicElement:
        return draw.Rectangle(
            self.x,
            self.y - self.width / 2,
            self.length,
            self.width,
            fill="black",
        )

    def draw_mpl(self, axes: plt.Axes) -> None:
        axes.add_patch(
            patches.Rectangle(
                (self.x, self.y - self.width / 2),
                self.length,
                self.width,
                facecolor="black",
            )
        )


@dataclass(slots=True, kw_only=True)
class PhaseShifterDrawing(DrawSpec):
    """
    Contains components for drawing a phase shifter using the different
    visualization methods.
    """

    x: float
    y: float
    size: float
    phase: str | float

    def draw_svg(self) -> draw.Group:
        g = draw.Group()
        g.append(
            CurvedRectangle(
                x=self.x,
                y=self.y - self.size / 2,
                size_x=self.size,
                size_y=self.size,
                radius=5,
                colour="#e8532b",
                outline="black",
            ).draw_svg()
        )
        g.append(
            TextDrawing(
                text="PS",
                x=self.x + self.size / 2,
                y=self.y + 2,
                rotation=0,
                size=SVGSettings.TEXT_SIZE.value,
                colour="white",
                alignment="centred",
            ).draw_svg()
        )
        phi_text = simplify_phase(self.phase)
        g.append(
            TextDrawing(
                text=f"φ = {phi_text}",
                x=self.x + self.size / 2,
                y=self.y + self.size,
                rotation=0,
                size=SVGSettings.S_TEXT_SIZE.value,
                colour="black",
                alignment="centred",
            ).draw_svg()
        )
        return g

    def draw_mpl(self, axes: plt.Axes) -> None:
        CurvedRectangle(
            x=self.x,
            y=self.y - self.size / 2,
            size_x=self.size,
            size_y=self.size,
            radius=0.07,
            colour="#e8532b",
            outline="black",
        ).draw_mpl(axes)
        TextDrawing(
            text="PS",
            x=self.x + self.size / 2,
            y=self.y,
            rotation=0,
            size=MPLSettings.TEXT_SIZE.value,
            colour="white",
            alignment="centred",
        ).draw_mpl(axes)
        phi_text = simplify_phase(self.phase)
        TextDrawing(
            text=f"$\\phi = {phi_text}$",
            x=self.x + self.size / 2,
            y=self.y + self.size / 2 + 0.15,
            rotation=0,
            size=MPLSettings.S_TEXT_SIZE.value,
            colour="black",
            alignment="centred",
        ).draw_mpl(axes)


@dataclass(slots=True, kw_only=True)
class BeamSplitterDrawing(DrawSpec):
    """
    Contains components for drawing a beam splitter using the different
    visualization methods.
    """

    x: float
    y: float
    size_x: float
    size_y: float
    offset_y: float
    reflectivity: str | float
    text_offset: float

    def __post_init__(self) -> None:
        if not isinstance(self.reflectivity, str):
            self.reflectivity = round(self.reflectivity, 4)

    def draw_svg(self) -> draw.Group:
        g = draw.Group()
        g.append(
            CurvedRectangle(
                x=self.x,
                y=self.y - self.offset_y,
                size_x=self.size_x,
                size_y=self.size_y,
                radius=5,
                colour="#3e368d",
                outline="black",
            ).draw_svg()
        )
        g.append(
            TextDrawing(
                text="BS",
                x=self.x + self.size_x / 2,
                y=(self.y + self.size_y / 2 - self.offset_y + self.text_offset),
                rotation=0,
                size=SVGSettings.TEXT_SIZE.value,
                colour="white",
                alignment="centred",
            ).draw_svg()
        )
        g.append(
            TextDrawing(
                text=f"r = {self.reflectivity}",
                x=self.x + self.size_x / 2,
                y=self.y + self.size_y,
                rotation=0,
                size=SVGSettings.S_TEXT_SIZE.value,
                colour="black",
                alignment="centred",
            ).draw_svg()
        )
        return g

    def draw_mpl(self, axes: plt.Axes) -> None:
        CurvedRectangle(
            x=self.x,
            y=self.y - self.offset_y / 2,
            size_x=self.size_x,
            size_y=self.size_y,
            radius=0.07,
            colour="#3e368d",
            outline="black",
        ).draw_mpl(axes)
        TextDrawing(
            text="BS",
            x=self.x + self.size_x / 2,
            y=self.y + 0.5,
            rotation=0,
            size=MPLSettings.TEXT_SIZE.value,
            alignment="centred",
            colour="white",
        ).draw_mpl(axes)
        TextDrawing(
            text=f"$r = ${self.reflectivity}",
            x=self.x + self.size_x / 2,
            y=self.y + self.size_y - self.offset_y / 2 + 0.15,
            rotation=0,
            size=MPLSettings.S_TEXT_SIZE.value,
            alignment="centred",
            colour="black",
        ).draw_mpl(axes)


@dataclass(slots=True, kw_only=True)
class UnitaryDrawing(DrawSpec):
    """
    Contains components for drawing a unitary using the different
    visualization methods.
    """

    x: float
    y: float
    size_x: float
    size_y: float
    offset_y: float
    label: str
    text_size: float

    def draw_svg(self) -> draw.Group:
        g = draw.Group()
        g.append(
            CurvedRectangle(
                x=self.x,
                y=self.y - self.offset_y,
                size_x=self.size_x,
                size_y=self.size_y,
                radius=5,
                colour="#1a0f36",
                outline="black",
            ).draw_svg()
        )
        s = self.text_size * 7 / 5 if len(self.label) == 1 else self.text_size
        r = 270 if len(self.label) > 2 else 0
        g.append(
            TextDrawing(
                text=self.label,
                x=self.x + self.size_x / 2,
                y=self.y + self.size_y / 2 - self.offset_y,
                rotation=r,
                size=s,
                colour="white",
                alignment="centred",
            ).draw_svg()
        )
        return g

    def draw_mpl(self, axes: plt.Axes) -> None:
        CurvedRectangle(
            x=self.x,
            y=self.y - self.offset_y / 2,
            size_x=self.size_x,
            size_y=self.size_y,
            radius=0.07,
            colour="#1a0f36",
            outline="black",
        ).draw_mpl(axes)
        s = 5 / 4 * self.text_size if len(self.label) == 1 else self.text_size
        r = 90 if len(self.label) > 2 else 0
        TextDrawing(
            text=self.label,
            x=self.x + self.size_x / 2,
            y=self.y + (self.size_y - self.offset_y) / 2,
            rotation=r,
            size=s,
            colour="white",
            alignment="centred",
        ).draw_mpl(axes)


@dataclass(slots=True, kw_only=True)
class LossDrawing(DrawSpec):
    """
    Contains components for drawing a loss element using the different
    visualization methods.
    """

    x: float
    y: float
    size: float
    loss: str | float

    def __post_init__(self) -> None:
        if not isinstance(self.loss, str):
            self.loss = str(round(self.loss, 4))

    def draw_svg(self) -> draw.Group:
        g = draw.Group()
        g.append(
            CurvedRectangle(
                x=self.x,
                y=self.y - self.size / 2,
                size_x=self.size,
                size_y=self.size,
                radius=5,
                colour="grey",
                outline="black",
            ).draw_svg()
        )
        g.append(
            TextDrawing(
                text="L",
                x=self.x + self.size / 2,
                y=self.y + 2,
                rotation=0,
                size=SVGSettings.TEXT_SIZE.value,
                colour="white",
                alignment="centred",
            ).draw_svg()
        )
        g.append(
            TextDrawing(
                text=f"loss = {self.loss}",
                x=self.x + self.size / 2,
                y=self.y + self.size,
                rotation=0,
                size=SVGSettings.S_TEXT_SIZE.value,
                colour="black",
                alignment="centred",
            ).draw_svg()
        )
        return g

    def draw_mpl(self, axes: plt.Axes) -> None:
        CurvedRectangle(
            x=self.x,
            y=self.y - self.size / 2,
            size_x=self.size,
            size_y=self.size,
            radius=0.07,
            colour="grey",
            outline="black",
        ).draw_mpl(axes)
        TextDrawing(
            text="L",
            x=self.x + self.size / 2,
            y=self.y,
            rotation=0,
            size=MPLSettings.TEXT_SIZE.value,
            colour="white",
            alignment="centred",
        ).draw_mpl(axes)
        TextDrawing(
            text=f"loss = ${self.loss}$",
            x=self.x + self.size / 2,
            y=self.y + self.size / 2 + 0.15,
            rotation=0,
            size=MPLSettings.S_TEXT_SIZE.value,
            colour="black",
            alignment="centred",
        ).draw_mpl(axes)


@dataclass(slots=True, kw_only=True)
class TextDrawing(DrawSpec):
    """
    Contains components for drawing text using the different visualization
    methods.
    """

    text: str
    x: float
    y: float
    rotation: float
    size: float
    colour: str
    alignment: Literal["left", "centred", "right"]

    def draw_svg(self) -> draw.DrawingBasicElement:
        if self.alignment == "centred":
            ta = "middle"
            db = "middle"
        elif self.alignment == "left":
            ta = "start"
            db = "middle"
        elif self.alignment == "right":
            ta = "end"
            db = "middle"
        else:
            raise DisplayError("Alignment value not recognised.")
        return draw.Text(
            self.text,
            self.size,
            self.x,
            self.y,
            fill=self.colour,
            text_anchor=ta,
            dominant_baseline=db,
            transform=f"rotate({self.rotation}, {self.x}, {self.y})",
        )

    def draw_mpl(self, axes: plt.Axes) -> None:
        if self.alignment == "centred":
            ha = "center"
            va = "center"
        elif self.alignment == "left":
            ha = "left"
            va = "center"
        elif self.alignment == "right":
            ha = "right"
            va = "center"
        else:
            raise DisplayError("Alignment value not recognised.")
        axes.text(
            self.x,
            self.y,
            self.text,
            horizontalalignment=ha,
            verticalalignment=va,
            color=self.colour,
            size=self.size,
            rotation=self.rotation,
        )


@dataclass(slots=True, kw_only=True)
class ModeSwapDrawing(DrawSpec):
    """
    Contains components for drawing mode swaps using the different
    visualization methods.
    """

    x: float
    ys: list[tuple[float, float]]
    size_x: float
    wg_width: float

    def draw_svg(self) -> draw.Group:
        g = draw.Group()
        for y0, y1 in self.ys:
            w = self.wg_width / 2
            m = np.arctan(abs(y1 - y0) / self.size_x)
            if y0 < y1:
                dx1 = w * m
                dx2 = 0
            else:
                dx1 = 0
                dx2 = w * m

            points = [
                self.x + dx1,
                y0 - w,
                self.x,
                y0 - w,
                self.x,
                y0 + w,
                self.x + dx2,
                y0 + w,
                self.x + self.size_x - dx1,
                y1 + w,
                self.x + self.size_x,
                y1 + w,
                self.x + self.size_x,
                y1 - w,
                self.x + self.size_x - dx2,
                y1 - w,
            ]
            poly = draw.Lines(*points, fill="black", close=True)
            g.append(poly)
        return g

    def draw_mpl(self, axes: plt.Axes) -> None:
        for y0, y1 in self.ys:
            w = self.wg_width / 2
            if y0 < y1:
                dx1 = w * np.arctan(abs(y1 - y0) / self.size_x)
                dx2 = 0
            else:
                dx1 = 0
                dx2 = w * np.arctan(abs(y1 - y0) / self.size_x)
            points = [
                (self.x + dx1, y0 - w),
                (self.x, y0 - w),
                (self.x, y0 + w),
                (self.x + dx2, y0 + w),
                (self.x + self.size_x - dx1, y1 + w),
                (self.x + self.size_x, y1 + w),
                (self.x + self.size_x, y1 - w),
                (self.x + self.size_x - dx2, y1 - w),
            ]
            axes.add_patch(patches.Polygon(points, facecolor="black"))


@dataclass(slots=True, kw_only=True)
class HeraldDrawing(DrawSpec):
    """
    Contains components for drawing herald markers using the different
    visualization methods.
    """

    x: float
    y: float
    size: float
    n_photons: int

    def draw_svg(self) -> draw.Group:
        g = draw.Group()
        g.append(
            draw.Circle(
                self.x, self.y, self.size, fill="#3e368d", stroke="black"
            )
        )
        g.append(
            TextDrawing(
                text=str(self.n_photons),
                x=self.x,
                y=self.y + 2.5,
                rotation=0,
                size=30,
                colour="white",
                alignment="centred",
            ).draw_svg()
        )
        return g

    def draw_mpl(self, axes: plt.Axes) -> None:
        axes.add_patch(
            patches.Circle(
                (self.x, self.y),
                self.size,
                facecolor="#3e368d",
                edgecolor="black",
            )
        )
        TextDrawing(
            text=str(self.n_photons),
            x=self.x,
            y=self.y + 0.01,
            rotation=0,
            size=MPLSettings.TEXT_SIZE.value,
            colour="white",
            alignment="centred",
        ).draw_mpl(axes)


@dataclass(slots=True, kw_only=True)
class BarrierDrawing(DrawSpec):
    """
    Contains components for drawing barriers markers using the different
    visualization methods.
    """

    x: float
    y_start: float
    y_end: float
    width: float

    def draw_svg(self) -> draw.DrawingBasicElement:
        return draw.Line(
            self.x,
            self.y_start,
            self.x,
            self.y_end,
            stroke="black",
            stroke_width=self.width,
        )

    def draw_mpl(self, axes: plt.Axes) -> None:
        axes.add_patch(
            patches.Rectangle(
                (self.x - self.width / 2, self.y_start),
                self.width,
                self.y_end - self.y_start,
                facecolor="black",
            )
        )


@dataclass(slots=True, kw_only=True)
class CurvedRectangle(DrawSpec):
    """
    Contains components for drawing curved rectangles using the different
    visualization methods.
    """

    x: float
    y: float
    size_x: float
    size_y: float
    radius: float
    colour: str
    outline: str

    def draw_svg(self) -> draw.DrawingBasicElement:
        return draw.Rectangle(
            self.x,
            self.y,
            self.size_x,
            self.size_y,
            fill=self.colour,
            stroke=self.outline,
            rx=self.radius,
            ry=self.radius,
        )

    def draw_mpl(self, axes: plt.Axes) -> None:
        axes.add_patch(
            patches.FancyBboxPatch(
                (self.x, self.y),
                self.size_x,
                self.size_y,
                boxstyle=f"Round, pad=0, rounding_size={self.radius}",
                facecolor=self.colour,
                edgecolor=self.outline,
            )
        )


def simplify_phase(phase: str | float) -> str:
    """
    Converts a phase values into a fraction of pi if it is some integer amount
    of π/4.
    """
    # Work out value of n*pi/4 closest to phi
    if not isinstance(phase, str):
        n = int(np.round(phase / (np.pi / 4)))
        # Check if value of phi == n*pi/4 to 8 decimal places
        if round(phase, 8) == round(n * np.pi / 4, 8):  # and n > 0:
            n = abs(n)
            # Set text with either pi or pi/2 or pi/4
            if n == 0:
                phi_text = "0"
            elif n % 4 == 0:
                phi_text = str(int(n / 4)) + "π" if n > 4 else "π"
            elif n % 4 == 2:
                phi_text = str(int(n / 2)) + "π/2" if n > 2 else "π/2"
            else:
                phi_text = str(int(n)) + "π/4" if n > 1 else "π/4"
            if phase < 0:
                phi_text = "-" + phi_text
            return phi_text
        # Otherwise round phi to 4 decimal places
        return str(round(phase, 4))
    return phase



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from types import FunctionType, MethodType
from typing import Any, overload

import matplotlib.figure
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from numpy.typing import NDArray

from lightworks.sdk.state import State
from lightworks.sdk.utils import PostSelectionFunction
from lightworks.sdk.utils.exceptions import ResultCreationError
from lightworks.sdk.utils.post_selection import PostSelectionType

from .result import Result


class SimulationResult(Result[State, dict[State, float | complex]]):
    """
    Stores results data from a given simulation in the emulator. There is then
    a range of options for displaying the data, or alternatively the data can
    be accessed directly using the [] operator on the class to select which
    input and output data is required.

    Args:

        results (np.ndarray) : The results which are to be stored.

        result_type (str) : The type of results which are being stored. This
            should either be probability, probability_amplitude or counts.

        inputs (list) : A list of the inputs used for creation of the results.

        outputs (list): A list of the possible outputs from the results.

    """

    def __init__(
        self,
        results: NDArray[np.float64 | np.complex128],
        result_type: str,
        inputs: list[State],
        outputs: list[State],
        **kwargs: Any,
    ) -> None:
        # Store result_type if valid
        if result_type in {"probability", "probability_amplitude"}:
            self.__result_type = result_type
        else:
            raise ResultCreationError(
                "Valid result type not provided, should either be "
                "'probability', 'probability_amplitude'."
            )

        self.__array = np.array(results)
        self.__inputs = inputs
        self.__outputs = outputs
        if len(self.__inputs) != self.__array.shape[0]:
            raise ResultCreationError(
                "Mismatch between inputs length and array size."
            )
        if len(self.__outputs) != self.__array.shape[1]:
            raise ResultCreationError(
                "Mismatch between outputs length and array size."
            )

        dict_results = {}
        for i, istate in enumerate(self.__inputs):
            input_results = {}
            for j, ostate in enumerate(self.__outputs):
                input_results[ostate] = self.__array[i, j]
            dict_results[istate] = input_results
        super().__init__(dict_results)

        # Store any additional provided data from kwargs as attributes
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def array(self) -> NDArray[np.float64 | np.complex128]:
        """
        The calculated array of data, where the first dimension corresponds to
        the inputs and the second dimension to the outputs.
        """
        return self.__array

    @property
    def inputs(self) -> list[State]:
        """All inputs which values were calculated for."""
        return self.__inputs

    @property
    def outputs(self) -> list[State]:
        """All outputs which values were calculated for."""
        return self.__outputs

    @property
    def result_type(self) -> str:
        """
        Details where the result is a probability or probability amplitude.
        """
        return self.__result_type

    @overload
    def __getitem__(self, item: State) -> dict[State, float | complex]: ...

    @overload
    def __getitem__(
        self, item: tuple[State]
    ) -> dict[State, float | complex]: ...

    @overload
    def __getitem__(self, item: tuple[State, State]) -> float | complex: ...

    def __getitem__(
        self, item: State | tuple[State] | tuple[State, State]
    ) -> float | complex | dict[State, float | complex]:
        """Custom get item behaviour - used when object accessed with []."""
        if isinstance(item, State):
            if item not in self:
                raise KeyError("Requested input state not in data.")
            return super().__getitem__(item)
        if isinstance(item, tuple):
            # Check only two values have been provided
            if len(item) > 2:
                raise ValueError(
                    "Get item can only contain a maximum of two values."
                )
            # Separate data into two states
            istate = item[0]
            ostate = item[1] if len(item) == 2 else None
            # Check all aspects are valid
            if not isinstance(istate, State) or not isinstance(
                ostate, State | type(None)
            ):
                raise TypeError("Get item values should have type State.")
            sub_r = self[istate]
            # If None provided as second value then return all results for input
            if ostate is None:
                return sub_r
            # Else return requested value
            if ostate not in sub_r:
                raise KeyError("Requested output state not in data.")
            return sub_r[ostate]
        raise TypeError("Get item value must be either one or two States.")

    def map(
        self,
        mapping: Callable[[State, Any], State],
        *args: Any,
        **kwargs: Any,
    ) -> "SimulationResult":
        """
        Performs a generic remapping of states based on a provided function.
        """
        if not isinstance(mapping, FunctionType | MethodType):
            raise TypeError(
                "Provided mapping should be a callable function which accepts "
                "and returns a State object."
            )
        if self.result_type == "probability_amplitude":
            raise ValueError(
                "Mapping cannot be applied to probability amplitudes."
            )
        mapped_result: dict[State, dict[State, float | complex]] = {}
        for in_state, results in self.items():
            mapped_result[in_state] = {}
            for out_state, val in results.items():
                new_s = mapping(out_state, *args, **kwargs)
                if new_s in mapped_result[in_state]:
                    mapped_result[in_state][new_s] += val
                else:
                    mapped_result[in_state][new_s] = val
        return self._recombine_mapped_result(mapped_result)

    def apply_post_selection(
        self, post_selection: PostSelectionType | Callable[[State], bool]
    ) -> "SimulationResult":
        """
        Applies an additional post-selection criteria to the stored result and
        returns this as a new object.
        """
        if isinstance(post_selection, FunctionType | MethodType):
            post_selection = PostSelectionFunction(post_selection)
        if not isinstance(post_selection, PostSelectionType):
            raise TypeError(
                "Provided post_selection should either be a PostSelection "
                "object or a callable function which accepts a state and "
                "returns a boolean to indicate whether the state is valid."
            )
        return self._recombine_mapped_result(
            {
                in_state: {
                    out_state: val
                    for out_state, val in results.items()
                    if post_selection.validate(out_state)
                }
                for in_state, results in self.items()
            }
        )

    def _recombine_mapped_result(
        self, mapped_result: dict[State, dict[State, float | complex]]
    ) -> "SimulationResult":
        """Creates a new Result object from mapped data."""
        unique_outputs: set[State] = set()
        for pdist in mapped_result.values():
            unique_outputs.update(pdist)
        array = np.zeros((len(self.inputs), len(unique_outputs)))
        for i, in_state in enumerate(self.inputs):
            for j, out_state in enumerate(unique_outputs):
                if out_state in mapped_result[in_state]:
                    array[i, j] = mapped_result[in_state][out_state]
        return SimulationResult(
            array,
            result_type=self.result_type,
            inputs=self.inputs,
            outputs=list(unique_outputs),
        )

    def plot(
        self,
        conv_to_probability: bool = False,
        show: bool = True,
        state_labels: dict[State, str | State] | None = None,
    ) -> tuple[matplotlib.figure.Figure, plt.Axes | NDArray[Any]] | None:
        """
        Create a plot of the data contained in the result. This will either
        take the form of a heatmap or bar chart, depending on the nature of the
        data contained in the Result object.

        Args:

            conv_to_probability (bool, optional) : In the case that the result
                is a probability amplitude, setting this to True will convert
                it into a probability.

            show (bool, optional) : Can be used to automatically show the
                created plot with show instead of returning the figure and
                axes.

            state_labels (dict, optional) : Provided a dictionary which can be
                used to specify alternate labels for each of the states when
                plotting. The keys of this dictionary should be States and the
                values should be strings or States.

        """
        if state_labels is None:
            state_labels = {}
        # Check provided state labels are valid
        for state, label in state_labels.items():
            if not isinstance(state, State):
                raise TypeError("Keys of state_labels dict should be States.")
            # Convert values from state_labels to strings if not already
            state_labels[state] = str(label)
        # Use imshow for any results originally provided as an array if more
        # than two more inputs are used.
        if len(self.inputs) >= 2:
            fig, ax = self._plot_array(conv_to_probability, state_labels)
        # Otherwise use a bar chart
        else:
            fig, ax = self._plot_bar(conv_to_probability, state_labels)
        # Optionally use show on plot if specified
        if show:
            plt.show()
            return None
        return (fig, ax)

    def print_outputs(self, rounding: int = 4) -> None:
        """
        Print the output results for each input into the system. This is
        compatible with all possible result types.

        Args:

            rounding (int, optional) : Set the number of decimal places which
                each number will be rounded to, defaults to 4.

        """
        # Loop over each input and print results
        for istate, results in self.items():
            to_print = str(istate) + " -> "
            for ostate, p in results.items():
                # Adjust print order based on quantity
                if self.result_type == "counts":
                    to_print += str(ostate) + " : " + str(p) + ", "
                else:
                    p = np.round(p, rounding)  # noqa: PLW2901
                    if abs(p.real) > 0 or abs(p.imag) > 0:
                        to_print += str(p) + "*" + str(ostate) + " + "
            to_print = to_print[:-2]
            print(to_print)  # noqa: T201

    def display_as_dataframe(
        self, threshold: float = 1e-12, conv_to_probability: bool = False
    ) -> pd.DataFrame:
        """
        Function to display the results of a given simulation in a dataframe
        format. Either the probability amplitudes of the state, or the actual
        probabilities can be displayed.

        Args:

            threshold (float, optional) : Threshold to control at which point
                value are rounded to zero. If looking for very small amplitudes
                this may need to be lowered.

            conv_to_probability (bool, optional) : In the case that the result
                is a probability amplitude, setting this to True will convert
                it into a probability. If it is not a probability amplitude
                then this setting will have no effect.

        Returns:

            pd.Dataframe : A dataframe with the results and input and output
                states as labels.

        """
        # Convert state vectors into strings
        in_strings = [str(s) for s in self.inputs]
        out_strings = [str(s) for s in self.outputs]
        # Switch to probability if required
        data = self.array.copy()
        if conv_to_probability and self.result_type == "probability_amplitude":
            data = abs(data) ** 2  # type: ignore[assignment]
        # Apply thresholding to values
        for i in range(data.shape[0]):
            for j in range(data.shape[1]):
                val = data[i, j]
                re = np.real(val) if abs(np.real(val)) > threshold else 0
                im = np.imag(val) if abs(np.imag(val)) > threshold else 0
                data[i, j] = re if abs(im) == 0 else re + 1j * im
        # Convert array to floats when not non complex results used
        if self.result_type != "probability_amplitude" or conv_to_probability:
            data = abs(data)
            if self.result_type == "counts" and not conv_to_probability:
                data = data.astype(int)
            else:
                data = data.astype(float)
        # Create dataframe and return
        return pd.DataFrame(data, index=in_strings, columns=out_strings)

    def _plot_array(
        self, conv_to_probability: bool, state_labels: dict[State, str | State]
    ) -> tuple[matplotlib.figure.Figure, plt.Axes | NDArray[Any]]:
        """
        Plots an array of the data contained within the Result object.
        """
        # Generate x and y labels
        xlabels = [
            state_labels[s] if s in state_labels else str(s)
            for s in self.outputs
        ]
        ylabels = [
            state_labels[s] if s in state_labels else str(s)
            for s in self.inputs
        ]
        # Single plots
        if self.result_type != "probability_amplitude" or conv_to_probability:
            if self.result_type == "probability_amplitude":
                a_data = abs(self.array) ** 2
            else:
                a_data = abs(self.array)
            # Plot array and add colorbar
            fig, ax = plt.subplots()
            im = ax.imshow(a_data)
            im_ratio = a_data.shape[0] / a_data.shape[1]
            fig.colorbar(im, ax=ax, fraction=0.046 * im_ratio, pad=0.04)
            # Label states on each axis
            ax.set_xticks(range(len(self.outputs)))
            ax.set_xticklabels(xlabels, rotation=90)  # type: ignore[arg-type]
            ax.set_yticks(range(len(self.inputs)))
            ax.set_yticklabels(ylabels)  # type: ignore[arg-type]
            return fig, ax
        # Otherwise create two plots
        fig, axes = plt.subplots(1, 2, figsize=(20, 6))
        axes = np.array(axes)
        vmin = min(op(self.array).min() for op in [np.real, np.imag])
        vmax = min(op(self.array).max() for op in [np.real, np.imag])
        im = axes[0].imshow(np.real(self.array), vmin=vmin, vmax=vmax)
        axes[0].set_title("real(result)")
        axes[1].imshow(np.imag(self.array), vmin=vmin, vmax=vmax)
        axes[1].set_title("imag(result)")
        for i in range(2):
            axes[i].set_xticks(range(len(self.outputs)))
            axes[i].set_xticklabels(xlabels, rotation=90)
            axes[i].set_yticks(range(len(self.inputs)))
            axes[i].set_yticklabels(ylabels)
        fig.colorbar(im, ax=axes.ravel().tolist())
        return (fig, axes)

    def _plot_bar(
        self, conv_to_probability: bool, state_labels: dict[State, str | State]
    ) -> tuple[matplotlib.figure.Figure, plt.Axes | NDArray[Any]]:
        """
        Plots bar chart with data contained in Results object. This should only
        be used if there is a single input.
        """
        istate = self.inputs[0]
        results = dict(self[istate])  # type: ignore[arg-type]
        # Vary plot depending on result type
        if self.result_type != "probability_amplitude" or conv_to_probability:
            d_data: dict[State, float]
            if self.result_type == "probability_amplitude":
                d_data = {}
                for s, p in results.items():
                    d_data[s] = abs(p) ** 2
                title = "Probability"
            else:
                d_data = results  # type: ignore[assignment]
                title = self.result_type.capitalize()

            fig, ax = plt.subplots(figsize=(7, 6))
            x_data = range(len(d_data))
            ax.bar(x_data, list(d_data.values()))
            ax.set_xticks(x_data)
            labels = [
                state_labels[s] if s in state_labels else str(s) for s in d_data
            ]
            ax.set_xticklabels(labels, rotation=90)  # type: ignore[arg-type]
            ax.set_xlabel("State")
            ax.set_ylabel(title)
            return (fig, ax)
        # Plot both real and imaginary parts
        fig, axes = plt.subplots(1, 2, figsize=(14, 6))
        axes = np.array(axes)
        x_data = range(len(results))
        axes[0].bar(x_data, np.real(list(results.values())))
        axes[1].bar(x_data, np.imag(list(results.values())))
        for i in range(2):
            axes[i].set_xticks(x_data)
            labels = [
                state_labels[s] if s in state_labels else str(s)
                for s in results
            ]
            axes[i].set_xticklabels(labels, rotation=90)
            axes[i].set_xlabel("State")
            axes[i].axhline(0, color="black", linewidth=0.5)
        axes[0].set_ylabel("real(amplitude)")
        axes[1].set_ylabel("imag(amplitude)")
        return (fig, axes)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any

import numpy as np
from multimethod import multimethod
from numpy.typing import NDArray

from lightworks.emulator.simulation import (
    AnalyzerRunner,
    SamplerRunner,
    SimulatorRunner,
)
from lightworks.emulator.utils.exceptions import BackendError
from lightworks.sdk.circuit.photonic_compiler import CompiledPhotonicCircuit
from lightworks.sdk.results import (
    ProbabilityDistribution,
    Result,
    SamplingResult,
    SimulationResult,
)
from lightworks.sdk.state import State
from lightworks.sdk.tasks import Analyzer, Sampler, Simulator, Task

from .abc_backend import EmulatorBackend

# ruff: noqa: ARG002, D102


class FockBackend(EmulatorBackend):
    """
    Base class for all backends. An outline of all possible functions should
    be included here.
    """

    @multimethod
    def run(self, task: Task) -> Result[State, Any]:
        raise BackendError("Task not supported on current backend.")

    @run.register
    def run_simulator(self, task: Simulator) -> SimulationResult:
        data = task._generate_task()
        return SimulatorRunner(
            data, self.probability_amplitude, self.state_generator
        ).run()

    @run.register
    def run_analyzer(self, task: Analyzer) -> SimulationResult:
        data = task._generate_task()
        return AnalyzerRunner(
            data, self.probability, self.state_generator
        ).run()

    @run.register
    def run_sampler(self, task: Sampler) -> SamplingResult:
        data = task._generate_task()
        runner = SamplerRunner(data, self.full_probability_distribution)
        cached_results = self._check_cache(data)
        if cached_results is not None:
            runner.probability_distribution = cached_results["pdist"]
            runner.herald_cache = cached_results["herald_cache"]
            task._probability_distribution = ProbabilityDistribution(
                cached_results["pdist"]
            )
        else:
            task._probability_distribution = ProbabilityDistribution(
                runner.distribution_calculator()
            )
            results = {
                "pdist": runner.probability_distribution,
                "herald_cache": runner.herald_cache,
            }
            self._add_to_cache(data, results)
        return runner.run()

    # Below defaults are defined for all possible methods in case they are
    # called without being implemented. This shouldn't normally happen.

    def state_generator(self, n_modes: int, n_photons: int) -> list[list[int]]:
        raise BackendError(
            "The required state generation function is not defined by this "
            "backend."
        )

    def probability_amplitude(
        self,
        unitary: NDArray[np.complex128],
        input_state: list[int],
        output_state: list[int],
    ) -> complex:
        raise BackendError(
            "Current backend does not implement probability_amplitude method."
        )

    def probability(
        self,
        unitary: NDArray[np.complex128],
        input_state: list[int],
        output_state: list[int],
    ) -> float:
        raise BackendError(
            "Current backend does not implement probability method."
        )

    def full_probability_distribution(
        self, circuit: CompiledPhotonicCircuit, input_state: State
    ) -> dict[State, float]:
        raise BackendError(
            "Current backend does not implement full_probability_distribution "
            "method."
        )



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from random import random

import pytest

from lightworks import State
from lightworks.emulator import Detector


class TestDetector:
    """
    Unit tests to check behaviour of the Detector object in different
    situations.
    """

    def setup_method(self) -> None:  # noqa: D102
        self.lossy_detector = Detector(efficiency=0.1)  # Should be very lossy
        self.dc_detector = Detector(p_dark=0.1)  # High probability
        self.non_pnr_detector = Detector(photon_counting=False)

    def test_threshold_detection_multiphoton(self):
        """
        Checks that threshold detection behaves as expected for a state with
        multiple photons in a mode.
        """
        out = self.non_pnr_detector._get_output(State([0, 1, 2, 3, 0, 1]))
        assert out == State([0, 1, 1, 1, 0, 1])

    def test_threshold_detection_singlephoton(self):
        """
        Checks that threshold detection behaves as expected for a state with
        all modes having one or less photons.
        """
        out = self.non_pnr_detector._get_output(State([0, 1, 1, 1, 0, 0]))
        assert out == State([0, 1, 1, 1, 0, 0])

    def test_threshold_detection_zerostate(self):
        """
        Checks that threshold detection behaves as expected for the zero state.
        """
        out = self.non_pnr_detector._get_output(State([0, 0, 0, 0, 0]))
        assert out == State([0, 0, 0, 0, 0])

    def test_lossy_detector(self):
        """
        Confirms that the behaviour of detectors with imperfect detection
        efficiency is as expected.
        """
        measured = set()
        measured.update(
            self.lossy_detector._get_output(State([0, 2, 1, 0]))
            for _ in range(1000)
        )
        assert State([0, 1, 0, 0]) in measured

    def test_dark_counts(self):
        """Test that dark counts are working as expected."""
        measured = set()
        measured.update(
            self.dc_detector._get_output(State([0, 0, 0, 0]))
            for _ in range(1000)
        )
        n_photons = [s.n_photons for s in measured]
        assert max(n_photons) > 0

    def test_efficiency_modification(self):
        """
        Checks that the efficiency cannot be assigned to an invalid value.
        """
        detector = Detector()
        with pytest.raises(TypeError):
            detector.efficiency = True
        with pytest.raises(ValueError):
            detector.efficiency = -0.1
        with pytest.raises(ValueError):
            detector.efficiency = 1.1

    def test_p_dark_modification(self):
        """
        Checks that the p_dark cannot be assigned to an invalid value.
        """
        detector = Detector()
        with pytest.raises(TypeError):
            detector.p_dark = True
        with pytest.raises(ValueError):
            detector.p_dark = -0.1
        with pytest.raises(ValueError):
            detector.p_dark = 1.1

    def test_photon_counting_modification(self):
        """
        Checks that the photon_counting cannot be assigned to an invalid value.
        """
        detector = Detector()
        with pytest.raises(TypeError):
            detector.photon_counting = 0.5
        with pytest.raises(TypeError):
            detector.photon_counting = "True"

    def test_detector_random_seeding(self):
        """
        Checks repeatable results are produced when random seeding is used.
        """
        r_seed = 100 * random()
        detector = Detector(efficiency=0.5, p_dark=1e-2)
        n_samples = 100
        input_state = State([1, 0, 2, 1, 0, 1, 0])
        # Get original results
        first = []
        detector._set_random_seed(r_seed)
        for _i in range(n_samples):
            first.append(detector._get_output(input_state))
        # Reset seed and sample again
        second = []
        detector._set_random_seed(r_seed)
        for _i in range(n_samples):
            second.append(detector._get_output(input_state))
        # Check they are equivalent
        assert first == second



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import inf

import pytest

from lightworks import (
    Parameter,
    ParameterBoundsError,
    ParameterDict,
    ParameterDictError,
    ParameterValueError,
)


class TestParameter:
    """Unit tests to test functionality of Parameter objects."""

    def test_parameter_creation(self):
        """
        Checks that a Parameter can be created as it's value can retrieved as
        expected.
        """
        p = Parameter(1)
        assert p.get() == 1

    def test_parameter_defaults(self):
        """
        Checks the default values for the different Parameter values when not
        specified.
        """
        p = Parameter(1)
        assert p.label is None
        assert p.min_bound is None
        assert p.max_bound is None

    def test_parameter_update(self):
        """Checks that a Parameter can be updated with set method."""
        p = Parameter(1)
        p.set(2)
        assert p.get() == 2

    def test_bounded_parameter_creation(self):
        """
        Confirms that a Parameter can be created with bounds and also checks an
        error is raised when set bounds don't contain Parameter value.
        """
        p = Parameter(1, bounds=[0, 2])
        assert p.min_bound == 0
        assert p.max_bound == 2
        with pytest.raises(ParameterBoundsError):
            p = Parameter(3, bounds=[0, 2])

    def test_bounded_parameter_update(self):
        """
        Checks that a bounded Parameter can be updated with set method, and
        also confirms an error is raised when parameters are set outside the
        bounds.
        """
        p = Parameter(1, bounds=[0, 3])
        p.set(2)
        assert p.get() == 2
        with pytest.raises(ParameterValueError):
            p.set(-1)
        with pytest.raises(ParameterValueError):
            p.set(4)

    def test_invalid_bound_update(self):
        """
        Checks that an error is raised when bounds are updated such that the
        current Parameter value would become invalid.
        """
        p = Parameter(2, bounds=[0, 3])
        with pytest.raises(ParameterBoundsError):
            p.max_bound = 1
        with pytest.raises(ParameterBoundsError):
            p.min_bound = 3

    def test_non_numerical_value(self):
        """
        Checks non-numerical values can be assigned for Parameter objects and
        also checks an error is raised if bounds are to be added.
        """
        p = Parameter(True)
        p.set(False)
        assert not p.get()
        with pytest.raises(ParameterBoundsError):
            p = Parameter(True, bounds=[0, 1])

    def test_non_numerical_bound_update(self):
        """
        Confirms that non-numerical Parameter cannot have bounds assigned after
        creation.
        """
        p = Parameter(True)
        with pytest.raises(ParameterBoundsError):
            p.min_bound = 0

    def test_bounded_value_type_updated(self):
        """
        Checks that a numerical Parameter with bounds cannot be switched to
        another non-numerical value.
        """
        p = Parameter(1, bounds=[0, 2])
        with pytest.raises(ParameterValueError):
            p.set("Test")

    def test_has_bounds_method(self):
        """Checks behaviour of the has_bounds method is correct."""
        p = Parameter(1)
        assert not p.has_bounds()
        p = Parameter(1, bounds=[0, 2])
        assert p.has_bounds()

    def test_str_return(self):
        """
        Confirms that Parameter value is returned when str function is used on
        Parameter.
        """
        p = Parameter(1.2)
        assert str(p) == "1.2"

    @pytest.mark.parametrize("value", [1, 2.1, "test"])
    def test_repr_return(self, value):
        """
        Checks Parameter value is contained within the value returned when repr
        is used.
        """
        p = Parameter(value)
        assert str(value) in repr(p)

    def test_label_in_repr_return(self):
        """Checks label included in repr return."""
        p = Parameter(1, label="label")
        assert "label" in repr(p)

    def test_bounds_in_repr_return(self):
        """Checks bounds in repr return."""
        p = Parameter(1, bounds=[0, 3])
        assert str(p.min_bound) in repr(p)
        assert str(p.max_bound) in repr(p)


class TestParameterDict:
    """Unit tests to test functionality of ParameterDict objects."""

    def test_dict_creation(self):
        """Checks dictionary can be created and is empty after creation."""
        pd = ParameterDict()
        assert pd.params == []

    def test_dict_assignment(self):
        """Check initial assignment of Parameter into dictionary."""
        pd = ParameterDict()
        pd["a"] = Parameter(1)
        assert pd["a"].get() == 1

    def test_dict_return(self):
        """Confirms that the return when using [] is a Parameter object."""
        pd = ParameterDict()
        pd["a"] = Parameter(1)
        assert isinstance(pd["a"], Parameter)

    def test_dict_invalid_param(self):
        """
        Confirms that a KeyError is raised when a parameter key does not exist.
        """
        pd = ParameterDict()
        pd["a"] = Parameter(1)
        with pytest.raises(KeyError):
            pd["b"]

    def test_parameter_update(self):
        """Test updating of parameter through assigning new value to dict."""
        pd = ParameterDict()
        pd["a"] = Parameter(1)
        pd["a"] = 2
        assert pd["a"].get() == 2

    def test_incorrect_set_item(self):
        """
        Checks an error is raised when a non-Parameter object is assigned to a
        new key and when it is attempted to overwrite a Parameter object in the
        dictionary with another Parameter.
        """
        pd = ParameterDict()
        with pytest.raises(ParameterDictError):
            pd["a"] = 1
        pd["a"] = Parameter(1)
        with pytest.raises(ParameterDictError):
            pd["a"] = Parameter(2)

    def test_len_method(self):
        """Confirms len method produces the correct value."""
        pd = ParameterDict()
        pd["a"] = Parameter(1)
        pd["b"] = Parameter(2)
        pd["c"] = Parameter(3)
        assert len(pd) == 3

    def test_has_bounds_method(self):
        """
        Checks has_bounds method returns correct values when bounds are/not
        present.
        """
        pd = ParameterDict()
        pd["a"] = Parameter(1)
        pd["b"] = Parameter(2)
        assert not pd.has_bounds()
        pd["a"].min_bound = 0
        assert pd.has_bounds()

    def test_params_property(self):
        """
        Confirms correct values are returned by params attribute of class.
        """
        pd = ParameterDict()
        pd["a"] = Parameter(1)
        pd["b"] = Parameter(2)
        assert pd.params == ["a", "b"]

    def test_params_modification(self):
        """Checks that params attribute cannot be modified."""
        pd = ParameterDict()
        with pytest.raises(AttributeError):
            pd.params = ["c"]

    def test_in_operator(self):
        """Tests that key in ParameterDict returns the correct values."""
        pd = ParameterDict()
        pd["a"] = Parameter(1)
        assert "a" in pd
        assert "b" not in pd
        assert "b" not in pd

    def test_remove_method(self):
        """
        Checks that remove method of ParameterDict is able to remove a
        parameter from the dictionary.
        """
        pd = ParameterDict()
        pd["a"] = Parameter(1)
        pd["b"] = Parameter(2)
        pd.remove("a")
        assert "a" not in pd

    def test_iterable_behavior(self):
        """
        Confirms that the values when using the class as an iterable is the
        parameter keys.
        """
        pd = ParameterDict()
        pd["a"] = Parameter(1)
        pd["b"] = Parameter(2)
        pd["c"] = Parameter(2)
        # Loop over dictionary and store returned values
        params = []
        for p in pd:
            params.append(p)
        # Check lists is equivalent to params attribute
        assert params == pd.params

    def test_get_bounds(self):
        """
        Confirms that get bounds is able to identify bounds for parameter.
        """
        pd = ParameterDict()
        pd["a"] = Parameter(1, bounds=[0, 2])
        assert pd.get_bounds()["a"] == (0, 2)

    def test_get_bounds_no_bounds(self):
        """
        Confirms that get bounds replaces None bounds with +/- inf.
        """
        pd = ParameterDict()
        pd["a"] = Parameter(1)
        assert pd.get_bounds()["a"] == (-inf, inf)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from lightworks.sdk.circuit import PhotonicCircuit
from lightworks.sdk.state import State
from lightworks.sdk.utils.state import check_herald_difference, validate_states

from .data import SimulatorTask
from .task import Task


class Simulator(Task):
    """
    Simulates a circuit for a provided number of inputs and outputs, returning
    the probability amplitude between them.

    Args:

        circuit (PhotonicCircuit) : The circuit which is to be used for
            simulation.

        inputs (list) : A list of the input states to simulate. For multiple
            inputs this should be a list of States.

        outputs (list | None, optional) : A list of the output states to
            simulate, this can also be set to None to automatically find all
            possible outputs.

    """

    def __init__(
        self,
        circuit: PhotonicCircuit,
        inputs: State | list[State],
        outputs: State | list[State] | None = None,
    ) -> None:
        # Assign circuit to attribute
        self.circuit = circuit
        self.inputs = inputs
        self.outputs = outputs

    @property
    def circuit(self) -> PhotonicCircuit:
        """
        Stores the circuit to be used for simulation, should be a
        PhotonicCircuit object.
        """
        return self.__circuit

    @circuit.setter
    def circuit(self, value: PhotonicCircuit) -> None:
        if not isinstance(value, PhotonicCircuit):
            raise TypeError(
                "Provided circuit should be a PhotonicCircuit or Unitary "
                "object."
            )
        self.__circuit = value

    @property
    def inputs(self) -> list[State]:
        """Stores a lost of all inputs to the system to simulate."""
        return self.__inputs

    @inputs.setter
    def inputs(self, value: State | list[State]) -> None:
        self.__inputs = validate_states(value, self.circuit.input_modes)

    @property
    def outputs(self) -> list[State] | None:
        """
        A list of all target outputs or None, if None all possible outputs are
        calculated on execution.
        """
        return self.__outputs

    @outputs.setter
    def outputs(self, value: State | list[State] | None) -> None:
        if value is not None:
            value = validate_states(value, self.circuit.input_modes)
        self.__outputs = value

    def _generate_task(self) -> SimulatorTask:
        check_herald_difference(self.circuit, self.inputs[0].n_photons)
        return SimulatorTask(
            circuit=self.circuit._build(),
            inputs=self.inputs,
            outputs=self.outputs,
        )



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import factorial
from typing import Any

import numpy as np
from numpy.typing import NDArray

from lightworks.__settings import settings
from lightworks.sdk.circuit.photonic_compiler import CompiledPhotonicCircuit
from lightworks.sdk.state import State

from .fock_backend import FockBackend


class SLOSBackend(FockBackend):
    """
    Contains calculate function for SLOS method.
    """

    @property
    def name(self) -> str:
        """Returns the name of the backend"""
        return "slos"

    @property
    def compatible_tasks(self) -> tuple[str, ...]:
        """Returns backends which are compatible with the backend."""
        return ("Sampler",)

    def full_probability_distribution(
        self, circuit: CompiledPhotonicCircuit, input_state: State
    ) -> dict[State, float]:
        """
        Finds the output probability distribution for the provided circuit and
        input state.

        Args:

            circuit (CompiledPhotonicCircuit) : The compiled version of the
                circuit which is being simulated. This is created by calling the
                _build method on the target circuit.

            input_state (State) : The input state to the system.

        Returns:

            dict : The calculated full probability distribution for the input.

        """
        # Return empty distribution when 0 photons in input
        if input_state.n_photons == 0:
            return {State([0] * circuit.n_modes): 1.0}

        pdist: dict[State, float] = {}
        # Add extra states for loss modes here when included
        if circuit.loss_modes > 0:
            input_state += State([0] * circuit.loss_modes)
        full_dist = self.calculate(circuit.U_full, input_state)
        # Combine results to remote lossy modes
        for s, p in full_dist.items():
            if abs(p) ** 2 > settings.sampler_probability_threshold:
                new_s = State(s[: circuit.n_modes])  # type: ignore[arg-type]
                if new_s in pdist:
                    pdist[new_s] += abs(p) ** 2
                else:
                    pdist[new_s] = abs(p) ** 2
        return pdist

    def calculate(
        self, unitary: NDArray[np.complex128], input_state: State
    ) -> dict[tuple[int, ...], complex]:
        """
        Performs calculation of full probability distribution given a unitary
        matrix and input state.
        """
        p = [m for m, n in enumerate(input_state) for _i in range(n)]
        n_modes = unitary.shape[0]
        # Normalise initial input probability
        m = 1 / np.sqrt(vector_factorial(input_state.s))
        input_ = {tuple(n_modes * [0]): m}  # N-mode vacuum state

        # Successively apply the matrices A_k
        for i in p:  # Each matrix is indexed by the components of p
            output: dict[tuple[int, ...], complex] = {}
            for j in range(n_modes):  # Sum over i
                step = a_i_dagger(
                    input_, j, unitary[j, i]
                )  # Apply ladder operator
                output = add_dicts(output, step)  # Add it to the total
            input_ = output

        return input_


def a_i_dagger(
    dist: dict[tuple[int, ...], complex], mode: int, multiplier: complex
) -> dict[tuple[int, ...], complex]:
    """
    Ladder operator for the ith mode applied to the state v, where v is a
    dictionary
    """
    updated_dist = {}  # Create a new dictionary to store updated values

    for key, value in dist.items():
        key = list(key)  # type: ignore[assignment] # noqa: PLW2901
        key[mode] += 1  # type: ignore[index]
        # Update the new dictionary with modified key, value + normalisation
        updated_dist[tuple(key)] = key[mode] ** 0.5 * value * multiplier

    return updated_dist


def vector_factorial(vector: list[int]) -> int:
    """Calculates the product of factorials of the elements of the vector v"""
    return int(np.prod([factorial(i) for i in vector]))


def add_dicts(dict1: dict[Any, Any], dict2: dict[Any, Any]) -> dict[Any, Any]:
    """Function for combining two dictionaries together"""
    for key, value in dict2.items():
        if key in dict1:
            dict1[key] += value
        else:
            dict1[key] = value
    return dict1



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ruff: noqa: PLW2901

"""
Contains a number of different utility functions for modifying circuits.
"""

from copy import copy
from numbers import Real
from typing import Any

import numpy as np
import sympy as sp
from multimethod import multimethod
from numpy.typing import NDArray

from lightworks.sdk.utils.param_unitary import ParameterizedUnitary

from .parameters import Parameter
from .photonic_components import (
    Barrier,
    BeamSplitter,
    Component,
    Group,
    HeraldData,
    Loss,
    ModeSwaps,
    PhaseShifter,
    UnitaryMatrix,
)


def unpack_circuit_spec(circuit_spec: list[Component]) -> list[Component]:
    """
    Unpacks and removes any grouped components from a circuit.

    Args:

        circuit_spec (list) : The circuit spec to unpack.

    Returns:

        list : The processed circuit spec.

    """
    new_spec = list(circuit_spec)
    while any(isinstance(s, Group) for s in new_spec):
        temp_spec = []
        for spec in circuit_spec:
            if not isinstance(spec, Group):
                temp_spec += [spec]
            else:
                temp_spec += spec.circuit_spec
        new_spec = temp_spec

    return new_spec


def add_modes_to_circuit_spec(
    circuit_spec: list[Component], mode: int
) -> list[Component]:
    """
    Takes an existing circuit spec and adds a given number of modes to each
    of the elements.

    Args:

        circuit_spec (list) : The circuit spec which is to be modified.

        mode (int) : The number of modes to shift each of the elements by.

    Returns:

        list : The modified version of the circuit spec.

    """
    new_circuit_spec = []
    for spec in circuit_spec:
        spec = copy(spec)
        if isinstance(spec, BeamSplitter):
            spec.mode_1 += mode
            spec.mode_2 += mode
        elif isinstance(spec, Barrier):
            spec.modes = [p + mode for p in spec.modes]
        elif isinstance(spec, ModeSwaps):
            spec.swaps = {k + mode: v + mode for k, v in spec.swaps.items()}
        elif isinstance(spec, Group):
            spec.circuit_spec = add_modes_to_circuit_spec(
                spec.circuit_spec, mode
            )
            spec.mode_1 += mode
            spec.mode_2 += mode
        else:
            spec.mode += mode  # type: ignore[attr-defined]
        new_circuit_spec.append(spec)
    return new_circuit_spec


def add_empty_mode_to_circuit_spec(
    circuit_spec: list[Component], mode: int
) -> list[Component]:
    """
    Takes a provided circuit spec and adds an empty mode at the set location.

    Args:

        circuit_spec (list) : The circuit spec which is to be modified.

        mode (int) : The location at which an empty mode should be included.

    Returns:

        list : The modified version of the circuit spec.

    """
    new_circuit_spec = []
    for spec in circuit_spec:
        spec = copy(spec)
        if isinstance(spec, BeamSplitter):
            spec.mode_1 += 1 if spec.mode_1 >= mode else 0
            spec.mode_2 += 1 if spec.mode_2 >= mode else 0
        elif isinstance(spec, Barrier):
            spec.modes = [p + 1 if p >= mode else p for p in spec.modes]
        elif isinstance(spec, ModeSwaps):
            swaps = {}
            for k, v in spec.swaps.items():
                k += 1 if k >= mode else 0
                v += 1 if v >= mode else 0
                swaps[k] = v
            spec.swaps = swaps
        elif isinstance(spec, Group):
            spec.circuit_spec = add_empty_mode_to_circuit_spec(
                spec.circuit_spec, mode
            )
            # Shift unitary mode range
            spec.mode_1 += 1 if spec.mode_1 >= mode else 0
            spec.mode_2 += 1 if spec.mode_2 >= mode else 0
            # Update herald values
            new_in_heralds = {}
            for m, n in spec.heralds.input.items():
                if m >= (mode - spec.mode_1) and mode - spec.mode_1 >= 0:
                    m += 1
                new_in_heralds[m] = n
            new_out_heralds = {}
            for m, n in spec.heralds.output.items():
                if m >= (mode - spec.mode_1) and mode - spec.mode_1 >= 0:
                    m += 1
                new_out_heralds[m] = n
            spec.heralds = HeraldData(
                input=new_in_heralds, output=new_out_heralds
            )
        elif isinstance(spec, UnitaryMatrix):
            spec.mode += 1 if spec.mode >= mode else 0
            # Expand unitary if required
            if spec.mode < mode < spec.mode + spec.unitary.shape[0]:
                add_mode = mode - spec.mode
                # Update unitary value
                spec.unitary = add_mode_to_unitary(spec.unitary, add_mode)
        else:
            spec.mode += 1 if spec.mode >= mode else 0  # type: ignore[attr-defined]
        new_circuit_spec.append(spec)
    return new_circuit_spec


def convert_non_adj_beamsplitters(
    circuit_spec: list[Component],
) -> list[Component]:
    """
    Takes a given circuit spec and removes all beam splitters acting on
    non-adjacent modes by replacing with a mode swap and adjacent beam
    splitters.

    Args:

        circuit_spec (list) : The circuit spec to remove beam splitter on
            non-adjacent modes from.

    Returns:

        list : The processed circuit spec.

    """
    new_spec: list[Component] = []
    for spec in circuit_spec:
        spec = copy(spec)
        if (
            isinstance(spec, BeamSplitter)
            and abs(spec.mode_2 - spec.mode_1) != 1
        ):
            m1, m2 = spec.mode_1, spec.mode_2
            if m1 > m2:
                m1, m2 = m2, m1
            mid = int((m1 + m2 - 1) / 2)
            swaps = {}
            for i in range(m1, mid + 1):
                swaps[i] = mid if i == m1 else i - 1
            for i in range(mid + 1, m2 + 1):
                swaps[i] = mid + 1 if i == m2 else i + 1
            new_spec.append(ModeSwaps(swaps))
            # If original modes were inverted then invert here too
            add1, add2 = mid, mid + 1
            if spec.mode_1 > spec.mode_2:
                add1, add2 = add2, add1
            # Add beam splitter on new modes
            new_spec.append(
                BeamSplitter(add1, add2, spec.reflectivity, spec.convention)
            )
            swaps = {v: k for k, v in swaps.items()}
            new_spec.append(ModeSwaps(swaps))
        elif isinstance(spec, Group):
            spec.circuit_spec = convert_non_adj_beamsplitters(spec.circuit_spec)
            new_spec.append(spec)
        else:
            new_spec.append(spec)
    return new_spec


def compress_mode_swaps(circuit_spec: list[Component]) -> list[Component]:
    """
    Takes a provided circuit spec and will try to compress any more swaps
    such that the circuit length is reduced. Note that any components in a
    group will be ignored.

    Args:

        circuit_spec (list) : The circuit spec which is to be processed.

    Returns:

        list : The processed version of the circuit spec.

    """
    new_spec: list[Component] = []
    to_skip: list[int] = []
    # Loop over each item in original spec
    for i, spec in enumerate(circuit_spec):
        spec = copy(spec)
        if i in to_skip:
            continue
        # If it a mode swap then check for subsequent mode swaps
        if isinstance(spec, ModeSwaps):
            blocked_modes = set()
            for j, spec2 in enumerate(circuit_spec[i + 1 :]):
                # Block modes with components other than the mode swap on
                if isinstance(spec2, PhaseShifter | Loss):
                    # NOTE: In principle a phase shift doesn't need to
                    # block a mode and instead we could modify it's
                    # location
                    blocked_modes.add(spec2.mode)
                elif isinstance(spec2, BeamSplitter):
                    blocked_modes.add(spec2.mode_1)
                    blocked_modes.add(spec2.mode_2)
                elif isinstance(spec2, Group):
                    blocked_modes.update(range(spec2.mode_1, spec2.mode_2 + 1))
                elif isinstance(spec2, UnitaryMatrix):
                    blocked_modes.update(
                        range(spec2.mode, spec2.mode + spec2.unitary.shape[0])
                    )
                elif isinstance(spec2, ModeSwaps):
                    # When a mode swap is found check if any of its mode
                    # are in the blocked mode
                    swaps = spec2.swaps
                    for m in swaps:
                        # If they are then block all other modes of swap
                        if m in blocked_modes:
                            blocked_modes.update(swaps)
                            break
                    else:
                        # Otherwise combine the original and found swap
                        # and update spec entry
                        new_swaps = combine_mode_swap_dicts(spec.swaps, swaps)
                        spec.swaps = new_swaps
                        # Also set to skip the swap that was combine
                        to_skip.append(i + 1 + j)
            new_spec.append(spec)
        else:
            new_spec.append(spec)

    return new_spec


def combine_mode_swap_dicts(
    swaps1: dict[int, int], swaps2: dict[int, int]
) -> dict[int, int]:
    """
    Function to take two mode swap dictionaries and combine them.

    Args:

        swaps1 (dict) : The first mode swap dictionary to combine.

        swaps2 (dict) : The mode swap dictionary to combine with the first
            dictionary.

    Returns:

        dict : The calculated combined mode swap dictionary.

    """
    # Store overall swaps in new dictionary
    new_swaps = {}
    added_swaps = []
    for k1, v1 in swaps1.items():
        for k2, v2 in swaps2.items():
            # Loop over swaps to combine when a key from swap 2 is in the
            # values of swap 1
            if v1 == k2:
                new_swaps[k1] = v2
                added_swaps.append(k2)
                break
        # If it isn't found then add key and value from swap 1
        else:
            new_swaps[k1] = v1
    # Add any keys from swaps2 that weren't used
    new_swaps.update({k: v for k, v in swaps2.items() if k not in added_swaps})
    # Remove any modes that are unchanged as these are not required
    return {m1: m2 for m1, m2 in new_swaps.items() if m1 != m2}


@multimethod
def check_loss(loss: Any) -> None:  # noqa: ARG001
    """
    Performs validation that a provided loss value is valid.
    """
    # Generic case to catch when an incorrect type is provided.
    raise TypeError(
        "Loss value should be a real numerical value or a Parameter. If set as "
        "a parameter then ensure the parameter value is a real number."
    )


@check_loss.register
def _check_loss_real(loss: Real) -> None:
    """
    Check that loss is assigned to a numerical value in the correct range.
    """
    if isinstance(loss, bool):
        raise TypeError("Loss value cannot be a boolean.")
    if not 0.0 <= float(loss) <= 1.0:
        raise ValueError("Provided loss values should be in the range [0,1].")


@check_loss.register
def _check_loss_param(loss: Parameter[Any]) -> None:
    """Check that loss value is valid when it is assigned to a parameter."""
    return check_loss(loss.get())


@multimethod
def add_mode_to_unitary(
    unitary: np.ndarray[Any, Any], add_mode: int
) -> NDArray[np.complex128]:
    """
    Adds a new mode (through inclusion of an extra row/column) to the provided
    unitary at the selected location.

    Args:

        unitary (np.ndarray | ParameterizedUnitary) : The unitary to add a mode
            to.

        add_mode (int) : The location at which a new mode should be added to
            the circuit.

    Returns:

        np.ndarray | ParameterizedUnitary : The converted unitary matrix.

    """
    dim = unitary.shape[0] + 1
    new_u = np.identity(dim, dtype=complex)
    # Diagonals
    new_u[:add_mode, :add_mode] = unitary[:add_mode, :add_mode]
    new_u[add_mode + 1 :, add_mode + 1 :] = unitary[add_mode:, add_mode:]
    # Off-diagonals
    new_u[:add_mode, add_mode + 1 :] = unitary[:add_mode, add_mode:]
    new_u[add_mode + 1 :, :add_mode] = unitary[add_mode:, :add_mode]
    return new_u


@add_mode_to_unitary.register
def _add_mode_to_parameterized_unitary(
    unitary: ParameterizedUnitary, add_mode: int
) -> ParameterizedUnitary:
    """Adds mode at selected location for ParameterizedUnitary."""
    dim = unitary.shape[0] + 1
    new_u = sp.Matrix([[float(i == j) for j in range(dim)] for i in range(dim)])
    # Diagonals
    new_u[:add_mode, :add_mode] = unitary._unitary[:add_mode, :add_mode]
    new_u[add_mode + 1 :, add_mode + 1 :] = unitary._unitary[
        add_mode:, add_mode:
    ]
    # Off-diagonals
    new_u[:add_mode, add_mode + 1 :] = unitary._unitary[:add_mode, add_mode:]
    new_u[add_mode + 1 :, :add_mode] = unitary._unitary[add_mode:, :add_mode]
    return ParameterizedUnitary(new_u, unitary._params)


def find_optimal_mode_swapping(
    required_swaps: dict[int, int], n_modes: int
) -> dict[int, int]:
    """
    Finds the optimal mode swapping (i.e. minimising swaps) for a particular set
    of required swaps by mapping to the first available mode.

    Args:

        required_swaps (dict) : The swaps that need to be realised by the
            operation.

        n_modes (int) : The number of modes across which swapping occurs.

    Returns:

        dict : The calculated swap dictionary.

    """
    current_mode = 0
    swaps = {}
    # Loop over all modes in circuit to find swaps
    for i in range(n_modes):
        # If used as a key then take value from provisional swaps
        if i in required_swaps:
            swaps[i] = required_swaps[i]
        # Otherwise then map mode to lowest mode possible
        else:
            while current_mode in required_swaps.values():
                current_mode += 1
            if i != current_mode:
                swaps[i] = current_mode
            current_mode += 1
    return swaps



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Literal, overload

import drawsvg
import matplotlib.figure
import matplotlib.pyplot as plt

from lightworks.sdk.utils.exceptions import DisplayError

from .draw_circuit_mpl import DrawCircuitMPL
from .draw_circuit_svg import DrawCircuitSVG

if TYPE_CHECKING:
    from lightworks.sdk.circuit import PhotonicCircuit


@overload
def display(
    circuit: "PhotonicCircuit",
    display_loss: bool = ...,
    mode_labels: list[str] | None = ...,
    display_type: Literal["svg"] = ...,
    show_parameter_values: bool = ...,
    display_barriers: bool = ...,
) -> drawsvg.Drawing: ...


@overload
def display(
    circuit: "PhotonicCircuit",
    display_loss: bool = ...,
    mode_labels: list[str] | None = ...,
    display_type: Literal["mpl"] = ...,
    show_parameter_values: bool = ...,
    display_barriers: bool = ...,
) -> tuple[matplotlib.figure.Figure, plt.Axes]: ...


# display function to interact with relevant classes
def display(
    circuit: "PhotonicCircuit",
    display_loss: bool = False,
    mode_labels: list[str] | None = None,
    display_type: Literal["svg", "mpl"] = "svg",
    show_parameter_values: bool = False,
    display_barriers: bool = False,
) -> tuple[matplotlib.figure.Figure, plt.Axes] | drawsvg.Drawing:
    """
    Used to display a circuit from lightworks in the chosen format.

    Args:

        circuit (PhotonicCircuit) : The circuit which is to be displayed.

        display_loss (bool, optional) : Choose whether to display loss
            components in the figure, defaults to False.

        mode_labels (list|None, optional) : Optionally provided a list of mode
            labels which will be used to name the mode something other than
            numerical values. Can be set to None to use default values.

        display_type (str, optional) : Selects whether the drawsvg or
            matplotlib module should be used for displaying the circuit. Should
            either be 'svg' or 'mpl', defaults to 'svg'.

        show_parameter_values (bool, optional) : Shows the values of parameters
            instead of the associated labels if specified.

        display_barriers (bool, optional) : Shows included barriers within the
            created visualization if this is required.

    Returns:

        (fig, ax) | Drawing : The created figure and axis or drawing for the
            display plot, depending on which display type is used.

    Raised:

        DisplayError : In any cases where an invalid display type is provided
            an exception will be raised.

    """
    if display_type == "mpl":
        disp = DrawCircuitMPL(
            circuit,
            display_loss,
            mode_labels,
            show_parameter_values,
            display_barriers,
        )
        return disp.draw()
    if display_type == "svg":
        disp_svg = DrawCircuitSVG(
            circuit,
            display_loss,
            mode_labels,
            show_parameter_values,
            display_barriers,
        )
        return disp_svg.draw()
    raise DisplayError("Display type not recognised.")



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from qiskit import QuantumCircuit


def convert_two_qubits_to_adjacent(
    q0: int, q1: int
) -> tuple[int, int, list[tuple[int, int]]]:
    """
    Takes two qubit indices and converts these so that they are adjacent to each
    other, and determining the swaps required for this. The order of the two
    qubits is preserved, so if q0 > q1 then this will remain True.

    Args:

        q0 (int) : First qubit which a gate acts on.

        q1 (int) : The second qubit which the gate acts on.

    Returns:

        int : The new first qubit which the gate should act on.

        int : The new second qubit which the gate should act on.

        list[tuple] : Pairs of qubits which swap gates should be applied to
            ensure the gate can act on the right qubits.

    """
    if abs(q1 - q0) == 1:
        return (q0, q1, [])
    swaps = []
    new_upper = max(q0, q1)
    new_lower = min(q0, q1)
    # Bring modes closer together until they are adjacent
    while new_upper - new_lower != 1:
        new_upper -= 1
        if new_upper - new_lower == 1:
            break
        new_lower += 1
    if min(q0, q1) != new_lower:
        swaps.append((min(q0, q1), new_lower))
    if max(q0, q1) != new_upper:
        swaps.append((max(q0, q1), new_upper))
    if q0 < q1:
        q0, q1 = new_lower, new_upper
    else:
        q0, q1 = new_upper, new_lower
    return (q0, q1, swaps)


def post_selection_analyzer(
    qc: "QuantumCircuit",
) -> tuple[list[bool], list[int]]:
    """
    Implements a basic algorithm to try to determine which gates can have
    post-selection and which require heralding. This is not necessarily optimal,
    but should at least reduce heralding requirements.

    Args:

        qc (QuantumCircuit) : The qiskit circuit to be analysed.

    Returns:

        list[bool] : A list of length elements in the circuit that indicates if
            each element is compatible with post-selection. This will include
            any single qubit gates, even though post-selection is not relevant
            here.

        list[int] : A list of integers indicating which qubits need a
            post-selection rule to be applied.

    """
    # First extract all qubit data from the circuit
    gate_qubits: list[list[int] | None] = []
    for inst in qc.data:
        if inst.operation.num_qubits >= 2 and inst.operation.name not in {
            "barrier",
            "swap",
        }:
            gate_qubits.append(
                [
                    inst.qubits[i]._index
                    for i in range(inst.operation.num_qubits)
                ]
            )
        else:
            gate_qubits.append(None)

    post_selection = []
    has_ps = []
    # Work backwards through gates
    for gate in reversed(gate_qubits):
        if gate is None:
            post_selection.append(False)
            continue
        can_ps = not all(q in has_ps for q in gate)
        post_selection.append(can_ps)
        has_ps += gate
    # Return if a gate can have post-selection and all modes which will require
    # it.
    return list(reversed(post_selection)), list(set(has_ps))



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.typing import NDArray

from lightworks.sdk.circuit import PhotonicCircuit
from lightworks.sdk.state import State

from .experiments import StateTomographyExperiment, StateTomographyList
from .mappings import MEASUREMENT_MAPPING
from .projection import _find_eigen_data
from .tomography import _Tomography
from .utils import (
    TomographyDataError,
    _calculate_density_matrix,
    _get_required_tomo_measurements,
    _get_tomo_measurements,
    state_fidelity,
)


class StateTomography(_Tomography):
    """
    Generates the required circuit and performs data processing for the
    calculation of the density matrix of a state.

    Args:

        n_qubits (int) : The number of qubits that will be used as part of the
            tomography.

        base_circuit (PhotonicCircuit) : An initial circuit which produces the
            required output state and can be modified for performing tomography.
            It is required that the number of circuit input modes equals 2 * the
            number of qubits.

    """

    @property
    def rho(self) -> NDArray[np.complex128]:
        """
        The most recently calculated density matrix.
        """
        if not hasattr(self, "_rho"):
            raise AttributeError(
                "Density matrix has not yet been calculated, this can be "
                "achieved with the process method."
            )
        return self._rho

    def get_experiments(self) -> StateTomographyList:
        """
        Generates all required tomography experiments for performing a process
        tomography algorithm.
        """
        req_measurements, _ = _get_required_tomo_measurements(self.n_qubits)

        # Generate all circuits and run experiment
        experiments = StateTomographyList()
        for gates in req_measurements:
            experiments.append(
                StateTomographyExperiment(
                    circuit=self._create_circuit(
                        [MEASUREMENT_MAPPING[g] for g in gates.split(",")]
                    ),
                    measurement_basis=gates,
                )
            )

        return experiments

    def process(
        self,
        data: list[dict[State, int]] | dict[str, dict[State, int]],
        project_to_physical: bool = False,
    ) -> NDArray[np.complex128]:
        """
        Performs the state tomography process with the configured elements to
        calculate the density matrix of the output state.

        Args:

            data (list | dict) : The collected measurement data. If a list then
                this should match the order the experiments were provided, and
                if a dictionary, then each key should be the corresponding
                measurement basis.

            project_to_physical (bool) : Controls whether the calculated density
                matrix is projected to a physical space. Defaults to False.

        Returns:

            np.ndarray : The calculated density matrix from the state tomography
                process.

        """
        req_measurements, result_mapping = _get_required_tomo_measurements(
            self.n_qubits
        )
        # Convert results into dictionary and then mapping to full set of
        # measurements
        if not isinstance(data, dict):
            if len(data) != len(req_measurements):
                msg = (
                    f"Number of results ({len(data)}) did not match the "
                    f"expected number ({len(req_measurements)}) for the target "
                    "tomography algorithm."
                )
                raise TomographyDataError(msg)
            results_dict = dict(zip(req_measurements, data, strict=True))
        else:
            missing = [meas for meas in req_measurements if meas not in data]
            if missing:
                msg = (
                    "One or more expected keys were detected to be missing "
                    f"from the results dictionary. Missing keys were {missing}."
                )
                raise TomographyDataError(msg)
            results_dict = dict(data)
        results_dict = {
            c: results_dict[result_mapping[c]]
            for c in _get_tomo_measurements(self.n_qubits)
        }

        self._rho = _calculate_density_matrix(
            results_dict, self.n_qubits, project_to_physical
        )
        return self.rho

    def fidelity(self, rho_exp: NDArray[np.complex128]) -> float:
        """
        Calculates the fidelity of the calculated quantum state against the
        expected density matrix for the state.

        Args:

            rho_exp (np.ndarray) : The expected density matrix.

        Returns:

            float : The calculated fidelity value.

        """
        return state_fidelity(self.rho, rho_exp)

    def check_eigenvalues(self) -> NDArray[np.float64]:
        """
        Determines the eigenvalues of the calculated density matrix, if the
        matrix is physical then these should all be non-negative.

        Returns:

            np.ndarray : An array of the calculated eigenvalues, ranging from
                smallest to largest.

        """
        return _find_eigen_data(self.rho)[0]

    def _create_circuit(
        self, measurement_operators: list[PhotonicCircuit]
    ) -> PhotonicCircuit:
        """
        Creates a copy of the assigned base circuit and applies the list of
        measurement circuits to each pair of dual-rail encoded qubits.

        Args:

            measurement_operators (list) : A list of 2 mode circuits which act
                as measurement operators to apply to the system.

        Returns:

            PhotonicCircuit : A modified copy of the base circuit with required
                operations.

        """
        circuit = self.base_circuit.copy()
        # Check number of circuits is correct
        if len(measurement_operators) != self.n_qubits:
            msg = (
                "Number of operators should match number of qubits "
                f"({self.n_qubits})."
            )
            raise ValueError(msg)
        # Add each and then return
        for i, op in enumerate(measurement_operators):
            circuit.add(op, 2 * i)

        return circuit
